#!/usr/bin/env python3
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum

import numpy as np
from scipy.spatial import distance
from skimage import feature, measure

from fibsem import conversions
from fibsem.imaging import masks
from fibsem.microscope import FibsemMicroscope
from fibsem.segmentation.model import SegmentationModel
from fibsem.structures import FibsemImage, BeamType, MicroscopeSettings, Point
import matplotlib.pyplot as plt

@dataclass
class Feature(ABC):
    px: Point 
    feature_m: Point
    color = "white"
    name: str = None

    @abstractmethod
    def detect(self, img: np.ndarray, mask: np.ndarray=None, point:Point=None) -> 'Feature':
        pass

    def to_dict(self):
        return {
            "name": self.name,
            "px": self.px.to_dict(),
            "feature_m": self.feature_m.to_dict()
        }

@dataclass
class ImageCentre(Feature):
    feature_m: Point = None
    px: Point = None
    color = "white"
    name: str = "ImageCentre"

    def detect(self, img: np.ndarray, mask: np.ndarray=None, point:Point=None) -> 'ImageCentre':
        self.px = Point(x=img.shape[1] // 2, y=img.shape[0] // 2)
        return self.px



@dataclass
class NeedleTip(Feature):
    feature_m: Point = None
    px: Point = None
    color = "green"
    name: str = "NeedleTip"
    class_id: int = 2
    class_name: str = "manipulator"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'NeedleTip':
        self.px = detect_needle_v5(mask, idx=self.class_id, edge="right")
        return self.px

@dataclass
class NeedleTipBottom(Feature):
    feature_m: Point = None
    px: Point = None
    color = "green"
    name: str = "NeedleTipBottom"
    class_id: int = 2
    class_name: str = "manipulator"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'NeedleTip':
        self.px = detect_needle_v5(mask, idx=self.class_id, edge="bottom")
        return self.px


@dataclass
class LamellaCentre(Feature):
    feature_m: Point = None
    px: Point = None
    color = "red"
    name: str = "LamellaCentre"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LamellaCentre':
        self.px = detect_lamella(mask, self)
        return self.px


@dataclass
class LamellaLeftEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "red"
    name: str = "LamellaLeftEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LamellaLeftEdge':
        self.px = detect_lamella(mask, self)
        return self.px


@dataclass
class LamellaRightEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "red"
    name: str = "LamellaRightEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LamellaRightEdge':
        self.px = detect_lamella(mask, self)
        return self.px

@dataclass
class LamellaTopEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "red"
    name: str = "LamellaTopEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LamellaTopEdge':
        self.px = detect_lamella(mask, self)
        return self.px

@dataclass
class LamellaBottomEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "red"
    name: str = "LamellaBottomEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LamellaBottomEdge':
        self.px = detect_lamella(mask, self)
        return self.px

@dataclass
class LandingPost(Feature):
    feature_m: Point = None
    px: Point = None
    color = "cyan"
    name: str = "LandingPost"
    class_id: int = 3
    class_name: str = "landing_post"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LandingPost':
        self.px = detect_landing_post_v4(mask, idx=self.class_id, point=point)
        # self.px = detect_landing_post_v3(img, landing_pt=None)
        return self.px


@dataclass
class LandingGridCentre(Feature):
    feature_m: Point = None
    px: Point = None
    color = "cyan"
    name: str = "LandingGridCentre"
    class_id: int = 3
    class_name: str = "landing_post"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'LandingGridCentre':
        mask = mask == self.class_id  
        self.px = detect_centre_point(mask, threshold=500)
        return self.px


@dataclass
class CoreFeature(Feature):
    feature_m: Point = None
    px: Point = None
    color = "white"
    name: str = "CoreFeature"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'CoreFeature':
        self.px = detect_core_feature(mask, self)
        return self.px

@dataclass
class CopperAdapterCentre(Feature):
    feature_m: Point = None
    px: Point = None
    color = "gold"
    name: str = "CopperAdapterCentre"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'CopperAdapterCentre':
        self.px = detect_centre_point(mask == 4)
        return self.px

@dataclass
class CopperAdapterTopEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "gold"
    name: str = "CopperAdapterTopEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'CopperAdapterTopEdge':
        self.px = detect_median_edge(mask == 4, edge="top")
        return self.px


@dataclass
class CopperAdapterBottomEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "gold"
    name: str = "CopperAdapterBottomEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'CopperAdapterBottomEdge':
        self.px = detect_median_edge(mask == 4, edge="bottom")
        return self.px


@dataclass
class VolumeBlockCentre(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockCentre"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockCentre':
        self.px = detect_centre_point(mask == 5)
        return self.px

@dataclass
class VolumeBlockTopEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockTopEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockTopEdge':
        self.px = detect_median_edge(mask == 5, edge="top")
        return self.px


@dataclass
class VolumeBlockBottomEdge(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockBottomEdge"

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockBottomEdge':
        self.px = detect_median_edge(mask == 5, edge="bottom")
        return self.px    


@dataclass
class VolumeBlockTopLeftCorner(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockTopLeftCorner"
    class_idx: int = 5

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockTopLeftCorner':
        try:    
            objects = get_objects(mask)
            objects = [obj for obj in objects if obj["class"] == self.class_idx]
            px = Point.from_list(objects[0]["keypoints"]["top_left"])
        except Exception as e:
            print(e)
            px = Point(0,0)
        self.px = px
        return self.px  

@dataclass
class VolumeBlockTopRightCorner(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockTopRightCorner"
    class_idx: int = 5

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockTopRightCorner':
        try:    
            objects = get_objects(mask)
            objects = [obj for obj in objects if obj["class"] == self.class_idx]
            px = Point.from_list(objects[0]["keypoints"]["top_right"])
        except Exception as e:
            print(e)
            px = Point(0,0)
        self.px = px
        return self.px 


@dataclass
class VolumeBlockBottomLeftCorner(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockBottomLeftCorner"
    class_idx: int = 5

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockBottomLeftCorner':
        try:    
            objects = get_objects(mask)
            objects = [obj for obj in objects if obj["class"] == self.class_idx]
            px = Point.from_list(objects[0]["keypoints"]["bottom_left"])
        except Exception as e:
            print(e)
            px = Point(0,0)
        self.px = px
        return self.px  

@dataclass
class VolumeBlockBottomRightCorner(Feature):
    feature_m: Point = None
    px: Point = None
    color = "magenta"
    name: str = "VolumeBlockBottomRightCorner"
    class_idx: int = 5

    def detect(self, img: np.ndarray, mask: np.ndarray = None, point:Point=None) -> 'VolumeBlockBottomRightCorner':
        try:    
            objects = get_objects(mask)
            objects = [obj for obj in objects if obj["class"] == self.class_idx]
            px = Point.from_list(objects[0]["keypoints"]["bottom_right"])
        except Exception as e:
            print(e)
            px = Point(0,0)
        self.px = px
        return self.px  

# TODO: we can probably consolidate this, rather than have so many classes
# Feature(class_idx, name, keypoint)
                
__FEATURES__ = [ImageCentre, NeedleTip, 
                LamellaCentre, LamellaLeftEdge, LamellaRightEdge, 
        LandingPost, LandingGridCentre, 
    CoreFeature, LamellaTopEdge, LamellaBottomEdge, 
    NeedleTipBottom, 
    CopperAdapterCentre, CopperAdapterTopEdge, CopperAdapterBottomEdge,
    VolumeBlockCentre, VolumeBlockTopEdge, VolumeBlockBottomEdge, 
    VolumeBlockTopLeftCorner, VolumeBlockTopRightCorner, VolumeBlockBottomLeftCorner, VolumeBlockBottomRightCorner ]
 


# Detection and Drawing Tools
def extract_class_pixels(mask, color):
    """ Extract only the pixels that are classified as the desired class (color)

    args:
        mask: detection mask containing all detection classes (np.array)
        color: the color of the specified class in the mask (rgb tuple)

    return:
        class_mask: the mask containing only the selected class
        idx: the indexes of the detected class in the mask

    """
    # TODO: should the class masks be binary?? probs easier

    # extract only label pixels to find edges
    class_mask = np.zeros_like(mask)
    idx = np.where(np.all(mask == color, axis=-1))
    class_mask[idx] = color

    return class_mask, idx


def detect_landing_post_v3(img: np.ndarray, landing_pt: Point = None, sigma=3) -> Point:
    if landing_pt is None:
        landing_pt = Point(x=img.shape[1] // 2, y=img.shape[0] // 2)
    edge = edge_detection(img, sigma=sigma)
    px = detect_closest_edge_v2(edge, landing_pt)
    return px

    
# TODO: generalise this to detect any edge
def detect_landing_post_v4(mask: np.ndarray, idx: int = 3, point: Point = None) -> Point:
    if point is None:
        point = Point(x=mask.shape[1] // 2, y=mask.shape[0] // 2)

    landing_mask = mask == idx

    # mask out outside 1/3
    idxs = int(landing_mask.shape[1] / 2.5)
    landing_mask[:, :idxs] = False
    landing_mask[:, -idxs:] = False

    # get median edge to
    # px = detect_median_edge(landing_mask, edge="top")

    px = detect_closest_edge_v2(landing_mask, point)
    return px

def detect_centre_point(mask: np.ndarray, threshold: int = 500) -> Point:
    """ Detect the centre (mean) point of the mask for a given color (label)

    args:
        mask: the detection mask (PIL.Image)
        idx: the index of the desired class in the mask (int)
        threshold: the minimum number of required pixels for a detection to count (int)

    return:

        centre_px: the pixel coordinates of the centre point of the feature (tuple)
    """
    centre_px = Point(x=0, y=0)
    # get mask px coordinates
    idx = np.where(mask)

    # only return a centre point if detection is above a threshold
    if len(idx[0]) > threshold:
        # get the centre point of each coordinate
        y_mid = int(np.mean(idx[0]))
        x_mid = int(np.mean(idx[1]))

        # centre coordinate as tuple
        centre_px = Point(x=x_mid, y=y_mid)
    return centre_px


def detect_corner(
    mask: np.ndarray, threshold=25, left: bool = False, bottom: bool = False
) -> Point:

    # get mask px coordinates
    edge_mask = np.where(mask)
    edge_px = (0, 0)

    # only return an edge point if detection is above a threshold
    if len(edge_mask[0]) > threshold:

        # right_most: max(x_coord), left_most: min(x_coord)
        px = np.max(edge_mask[1])
        if left:
            px = np.min(edge_mask[1])

        # get all px corresponding to value
        h_idx = np.where(edge_mask[1] == px)
        coords = edge_mask[0][h_idx], edge_mask[1][h_idx]

        # get the coordinates of the median vertical index
        py = int(np.median(coords[0]))
        edge_px = (py, px)

    return Point(x=int(edge_px[1]), y=int(edge_px[0]))


def detect_median_edge(mask: np.ndarray, edge: str, threshold: int = 250) -> Point:

    # get mask px coordindates
    edge_mask = np.where(mask)
    edge_px = Point(0, 0)

    if len(edge_mask[0]) > threshold:
        try:
            # get the centre point of each coordinate
            py = int(np.mean(edge_mask[0]))
            px = int(np.mean(edge_mask[1]))

            # get the min and max x and y coordinates at these median points
            x_min = np.min(edge_mask[1][edge_mask[0] == py])
            x_max = np.max(edge_mask[1][edge_mask[0] == py])
            y_min = np.min(edge_mask[0][edge_mask[1] == px])
            y_max = np.max(edge_mask[0][edge_mask[1] == px])


            if edge == "left":
                edge_px = Point(x=x_min, y=py)
            if edge == "right":
                edge_px = Point(x=x_max, y=py)
            if edge == "top":
                edge_px = Point(x=px, y=y_min)
            if edge == "bottom":
                edge_px = Point(x=px, y=y_max)
        except:
            pass
    return Point(x=int(edge_px.x), y=int(edge_px.y))


def detect_absolute_edge(mask, edge: str, _filter:str = "largest", _mode: str = "median", threshold: int = 150) -> Point:

    # _mode: median or strict
    #   median: take median of points with same extremity
    #   strict: first best selection

    # get individual objects
    labels = measure.label(mask)
    props = measure.regionprops(labels)

    if len(props) == 0:
        return Point(0,0)

    obj = props[0]
    if _filter == "largest":
        # get the largest object
        for prop in props:
            if prop.area > obj.area:
                obj = prop


    # get the coords of the most right pixel
    px = obj.coords[0]
    for coord in obj.coords:

        if edge == "right":
            if coord[1] > px[1]:
                px = coord
                
        
        if edge == "left":
            if coord[1] < px[1]:
                px = coord
        
        if edge == "top":
            if coord[0] < px[0]:
                px = coord
        
        if edge == "bottom":
            if coord[0] > px[0]:
                px = coord

    # find coordinates that have same x or y value as the most right pixel
    same_y = [coord for coord in obj.coords if coord[1] == px[1]]
    same_x = [coord for coord in obj.coords if coord[0] == px[0]]

    # TODO: could add some fuzz to the selection too (e..g +/- 2px)

    if edge in ["left", "right"]:
        # take median of points with same y value
        px = np.median(same_y, axis=0)
    else:
        # take median of points with same x value
        px = np.median(same_x, axis=0)

    return Point(px[1], px[0])

def detect_core_feature(
    mask: np.ndarray,
    feature: Feature,
    mask_radius: int = 512,
    idx: int = 1,
) -> Point:
    px = detect_centre_point(mask == idx, threshold=500)
    return px

def detect_lamella(
    mask: np.ndarray,
    feature: Feature,
    mask_radius: int = 512,
    idx: int = 1,
) -> Point:

    lamella_mask = mask == idx
    # lamella_mask = masks.apply_circular_mask(lamella_mask, radius=mask_radius)

    if isinstance(feature, LamellaCentre):
        px = detect_centre_point(lamella_mask)

    if isinstance(feature, LamellaLeftEdge):
        # px = detect_corner(lamella_mask, left=True)
        px = detect_median_edge(lamella_mask, edge="left")


    if isinstance(feature, LamellaRightEdge):
        # px = detect_corner(lamella_mask, left=False)
        px = detect_median_edge(lamella_mask, edge="right")


    if isinstance(feature, LamellaTopEdge):
        px = detect_median_edge(lamella_mask, edge="top")

    if isinstance(feature, LamellaBottomEdge):
        px = detect_median_edge(lamella_mask, edge="bottom")
    
    return px


def detect_needle_v4(mask: np.ndarray, idx:int=2) -> Point:
    needle_mask = mask == idx
    return detect_corner(needle_mask, threshold=100)

def detect_needle_v5(mask: np.ndarray, idx:int=2, edge: str ="right") -> Point:
    needle_mask = mask == idx
    return detect_absolute_edge(needle_mask, edge=edge, 
        _filter="largest", _mode="median", threshold=150)

def edge_detection(img: np.ndarray, sigma=3) -> np.ndarray:
    return feature.canny(img, sigma=sigma)  # sigma higher usually better


def detect_closest_edge_v2(
    mask: np.ndarray, landing_pt: Point
) -> tuple[Point, np.ndarray]:
    """ Identify the closest edge point to the initially selected point

    args:
        img: base image (np.ndarray)
        landing_px: the initial landing point pixel (tuple) (y, x) format
    return:
        landing_edge_pt: the closest edge point to the intitially selected point (tuple)
        edges: the edge mask (np.array)
    """

    # identify edge pixels
    landing_px = (landing_pt.y, landing_pt.x)
    edge_mask = np.where(mask)
    edge_px = list(zip(edge_mask[0], edge_mask[1]))

    # set min distance
    min_dst = np.inf

    # TODO: vectorise this like
    # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.euclidean_distances.html

    landing_edge_px = (0, 0)
    for px in edge_px:

        # distance between edges and landing point
        dst = distance.euclidean(landing_px, px)

        # select point with min
        if dst < min_dst:
            min_dst = dst
            landing_edge_px = px

    return Point(x=int(landing_edge_px[1]), y=int(landing_edge_px[0]))


def detect_bounding_box(mask, color, threshold=25):
    """ Detect the bounding edge points of the mask for a given color (label)

    args:
        mask: the detection mask (PIL.Image)
        color: the color of the label for the feature to detect (rgb tuple)
        threshold: the minimum number of required pixels for a detection to count (int)

    return:

        edge_px: the pixel coordinates of the edge points of the feature list((tuple))
    """

    top_px, bottom_px, left_px, right_px = (0, 0), (0, 0), (0, 0), (0, 0)

    # extract class pixels
    class_mask, idx = extract_class_pixels(mask, color)

    # only return an edge point if detection is above a threshold

    if len(idx[0]) > threshold:
        # convert mask to coordinates
        px = list(zip(idx[0], idx[1]))

        # get index of each value
        top_idx = np.argmin(idx[0])
        bottom_idx = np.argmax(idx[0])
        left_idx = np.argmin(idx[1])
        right_idx = np.argmax(idx[1])

        # pixel coordinates
        top_px = px[top_idx]
        bottom_px = px[bottom_idx]
        left_px = px[left_idx]
        right_px = px[right_idx]

    # bbox should be (x0, y0), (x1, y1)
    x0 = top_px[0]
    y0 = left_px[1]
    x1 = bottom_px[0]
    y1 = right_px[1]

    bbox = (x0, y0, x1, y1)

    return bbox

from typing import Union
### v2

@dataclass
class DetectedFeatures:
    features: list[Feature]
    image: np.ndarray # TODO: convert or add FIBSEMImage
    mask: np.ndarray # class binary mask
    rgb: np.ndarray # rgb mask
    pixelsize: float
    _distance: Point = None
    _offset: Point = Point(0, 0)
    fibsem_image: FibsemImage = None
    checkpoint: str = None

    @property
    def distance(self):
        assert len(self.features) >= 2, "Need at least two features to calculate distance"
        return self.features[0].px.distance(self.features[1].px)._to_metres(self.pixelsize) + self._offset        
        
    @distance.setter
    def distance(self, value: Point) -> None:
        self._distance = value

    def get_feature(self, ftype: Union[str, Feature]) -> Feature:
        for feature in self.features:
            if feature.name == ftype or isinstance(feature, type(ftype)):
                return feature

    def to_dict(self):
        return {
            "features": [f.to_dict() for f in self.features],
            "pixelsize": self.pixelsize,
            "offset": self._offset.to_dict(),
        }
        
def detect_features_v2(
    img: np.ndarray, mask: np.ndarray, features: tuple[Feature], filter: bool = True, point: Point = None
) -> list[Feature]:

    detection_features = []

    for feature in features:
        
        if isinstance(feature, (LamellaCentre, LamellaLeftEdge, LamellaRightEdge, LamellaTopEdge, LamellaBottomEdge, CoreFeature)):
            feature = detect_multi_features(img, mask, feature)
            if filter:
                feature = filter_best_feature(mask, feature, 
                                              method="closest", 
                                              point=point)
        else:
            feature.detect(img=img,mask=mask) 

        if isinstance(feature, list):
            detection_features.extend(feature)
        else:
            detection_features.append(feature)

        
    return detection_features


def detect_features(
    image: Union[np.ndarray, FibsemImage],
    model: SegmentationModel,
    features: tuple[Feature],
    pixelsize: float = None,
    filter: bool = True,
    point: Point = None
) -> DetectedFeatures:

    if isinstance(image, FibsemImage):
        fibsem_image = deepcopy(image)
        image = image.data
    else:
        fibsem_image = None

    if pixelsize is None:
        try:
            pixelsize = image.metadata.pixel_size.x
        except: # default (wrong value)
            pixelsize = 25e-9
        
    # model inference
    mask = model.inference(image, rgb=False)
    rgb = model.postprocess(mask, model.num_classes)
    mask = mask[0] # remove channel dim

    # detect features
    features = detect_features_v2(img=image, 
                                  mask=mask, 
                                  features=features, 
                                  filter=filter, point=point)

    det = DetectedFeatures(
        features=features, # type: ignore
        image=image,
        mask=mask,
        rgb=rgb,
        pixelsize=pixelsize,
        fibsem_image=fibsem_image,
        checkpoint=model.checkpoint
    )

    # distance in metres (from centre)
    for feature in det.features:
        feature.feature_m = conversions.image_to_microscope_image_coordinates(
            feature.px, det.image.data, det.pixelsize
        )

    return det


def take_image_and_detect_features(
    microscope: FibsemMicroscope,
    settings: MicroscopeSettings,
    features: tuple[Feature],
    point: Point = None,
) -> DetectedFeatures:
    
    from fibsem import acquire, utils
    from fibsem.segmentation.model import load_model
    from fibsem import config as cfg

    if settings.image.reduced_area is not None:
        logging.info(
            f"Reduced area is not compatible with model detection, disabling..."
        )
        settings.image.reduced_area = None
    
    settings.image.filename = f"ml-{utils.current_timestamp_v2()}"
    settings.image.save = True

    # take new image
    image = acquire.new_image(microscope, settings.image)

    # load model

    checkpoint = settings.protocol["options"].get("checkpoint", cfg.__DEFAULT_CHECKPOINT__)
    model = load_model(checkpoint=checkpoint)

    if isinstance(point, FibsemStagePosition):
        logging.debug(f"Reprojecting point {point} to image coordinates...")
        points = _tile._reproject_positions(image, [point], _bound=True)
        point = points[0] if len(points) == 1 else None
        logging.debug(f"Reprojected point: {point}")

    # detect features
    det = detect_features(
        deepcopy(image), model, features=features, pixelsize=image.metadata.pixel_size.x, point = point
    )
    return det



def plot_detection(det: DetectedFeatures):

    """Plotting image with detected features

    Args:
        det (DetectedFeatures): detected features type
        inverse (bool, optional): Inverses the colour of the centre crosshair of the feature. Defaults to True.
    """
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots(1, 1, figsize=(12, 7))

    fig = plot_det(det, ax)
    
    return fig


def plot_det(det: DetectedFeatures, ax: plt.Axes, title: str = "Prediction", show: bool = True):
    ax.imshow(det.image, cmap="gray")
    if det.rgb is not None:
        ax.imshow(det.rgb, alpha=0.3)
    ax.set_title(title)
    

    # get unique feature names
    names = []
    for f in det.features:
        if f.name not in names:
            names.append(f.name)
    

    for f in det.features:
        ax.plot(f.px.x, f.px.y, 
                "o",  color=f.color, 
                markersize=5, markeredgecolor="w", 
                label=f.name if names.count(f.name) == 1 else None)
        
        # remove from names list
        if names.count(f.name) == 1:
            names.remove(f.name)

    # if len(det.features) < 5:
    if len(det.features) != 0:
        ax.legend(loc="best")
    ax.axis("off")

    # if len(det.features) == 2:
    #     # plot white line between features
    #     ax.plot([det.features[0].px.x, det.features[1].px.x],
    #             [det.features[0].px.y, det.features[1].px.y], 
    #             color="w", linestyle="--")

    if show:
        plt.show()


def plot_detections(dets: list[DetectedFeatures], titles: list[str] = None) -> plt.Figure:
    """Plotting image with detected features

    Args:
        det (DetectedFeatures): detected features type
        inverse (bool, optional): Inverses the colour of the centre crosshair of the feature. Defaults to True.
    """
    import matplotlib.pyplot as plt

    if titles is None:
        titles = [f"Prediction {i}" for i in range(len(dets))]

    fig, ax = plt.subplots(1, len(dets), figsize=(25, 10))

    for i, det in enumerate(dets):
        if len(dets) == 1:
            fig = plot_det(det, ax, title=titles[i], show=False)
        else:
            plot_det(det, ax[i], title=titles[i], show=False)
    
    plt.subplots_adjust(wspace=0.05, hspace=0.05)
    # plt.show()

    return fig



_DETECTIONS_THAT_MOVE_MANIPULATOR = (NeedleTip, NeedleTipBottom, LamellaRightEdge, LamellaBottomEdge, LamellaTopEdge, CopperAdapterTopEdge, CopperAdapterBottomEdge)
_DETECTIONS_THAT_MOVE_STAGE = (LamellaCentre, LandingGridCentre)


# isinstance(feature, (LamellaCentre, LamellaLeftEdge, LamellaRightEdge, LamellaTopEdge, LamellaBottomEdge, CoreFeature)):


def move_based_on_detection(
    microscope: FibsemMicroscope,
    settings: MicroscopeSettings,
    det: DetectedFeatures,
    beam_type: BeamType,
    move_x: bool = True,
    move_y: bool = True,
    _move_system: str = None, # auto
):
    from fibsem import movement

    dx, dy = det.distance.x, det.distance.y

    # nulify movements in unsupported axes
    if not move_x:
        dx = 0
    if not move_y:
        dy = 0

    # f1 = move from, f2 = move to
    f1 = det.features[0]
    f2 = det.features[1]

    logging.debug(f"move_x: {move_x}, move_y: {move_y}")
    logging.debug(f"movement: x={dx:.2e}, y={dy:.2e}")
    logging.debug(f"features: {f1}, {f2}, beam_type: {beam_type}")

    if _move_system is None:
        if isinstance(f1, _DETECTIONS_THAT_MOVE_MANIPULATOR):
            _move_system = "manipulator"
        if isinstance(f1, _DETECTIONS_THAT_MOVE_STAGE):
            _move_system = "stage"

    if _move_system not in ["manipulator", "stage"]:
        raise ValueError(f"move_system must be one of ['manipulator', 'stage'], not {_move_system}")

    # these movements move the needle...
    if _move_system == "manipulator":

        # account for scan_rotation
        if np.isclose(microscope.get("scan_rotation", beam_type), np.pi):
            dx *= -1.0
            dy *= -1.0
        
        # NOTE (pat): double check this on liftout?
        if beam_type == BeamType.ELECTRON:
            dy *= -1.0
                
        microscope.move_manipulator_corrected(
            dx=dx,
            dy=dy,
            beam_type=beam_type,
        )

    if _move_system == "stage":
            # need to reverse the direction to move correctly. investigate if this is to do with scan rotation?
            microscope.stable_move(
                dx=-dx,
                dy=dy,
                beam_type=beam_type,
            )

            # TODO: support other movements?
    return


import glob
import os
from pathlib import Path
from typing import Optional

import pandas as pd
import tifffile as tff

from fibsem.segmentation import utils as seg_utils
from fibsem.structures import FibsemImage, Point, FibsemStagePosition


def mask_contours(image):
    # Find contours
    contours = measure.find_contours(image, 0.5)

    # Create a mask with the same shape as the input image
    mask = np.zeros_like(image, dtype=np.uint8)

    # Mask the area inside each contour
    for i, contour in enumerate(contours):
        # Convert the contour coordinates to integers
        contour = np.round(contour).astype(int)

        # Create a polygonal mask for this contour
        rr, cc = np.meshgrid(np.arange(mask.shape[0]), np.arange(mask.shape[1]), indexing='ij')
        inside_contour = np.zeros_like(mask, dtype=np.uint8)
        inside_contour[rr, cc] = 0
        inside_contour[contour[:, 0], contour[:, 1]] = i+1

        # mask the area inside the contour
        ymin, ymax = np.min(contour[:, 0]), np.max(contour[:, 0])
        xmin, xmax = np.min(contour[:, 1]), np.max(contour[:, 1])

        inside_contour[ymin:ymax, xmin:xmax] = i+1

        # Apply the mask for this contour
        mask = mask + inside_contour

    return mask

from copy import deepcopy
# TODO: need passthrough for the params
def detect_multi_features(image: np.ndarray, mask: np.ndarray, feature: Feature, class_idx: int = 1):
    
    mask = mask == class_idx # filter to class 
    mask = mask_contours(mask)
    idxs = np.unique(mask)

    features = []
    for idx in idxs:
        if idx==0:
            continue

        # create a new image
        feature_mask = np.zeros_like(mask)
        feature_mask[mask==idx] = 1

        # detect features
        feature.detect(image, feature_mask)
        features.append(deepcopy(feature))

    if features == []:
        logging.info(f"No features detected for {feature.name}")
        # set at centre of image
        feature.px = Point(x=image.shape[1]//2, y=image.shape[0]//2)
        features = [deepcopy(feature)]
    
    return features


def filter_best_feature(mask: np.ndarray, features: list[Feature], method: str = "closest", point: Point = None):
    if method == "closest":
        # plot feature closest to point
        if point is None:
            point = Point(mask.shape[1]/2, mask.shape[0]/2)

        distances = []
        for feature in features:
            distances.append(np.linalg.norm(point - feature.px))
        idx = np.argmin(distances)

        return features[idx]

    else:
        raise ValueError(f"method {method} not recognised")

def get_feature(name: str) -> Feature:

    idx = [i for i, feat in enumerate(__FEATURES__) if feat.__name__ == name][0]
    feature = __FEATURES__[idx]()

    return feature



# v4 intersection features
from fibsem.microscope import FibsemMicroscope 
from fibsem.structures import MicroscopeSettings
import numpy as np
from fibsem.imaging import _tile

def _detect_positions(microscope: FibsemMicroscope, settings: MicroscopeSettings, image: FibsemImage, mask:np.ndarray, features: list[Feature]) -> list[FibsemStagePosition]:

    # detect features
    features = detect_features_v2(img=image, 
                                mask=mask, 
                                features=features, 
                                filter=False, point=None)

    # convert image coordinates to microscope coordinates # TODO: check why we reverse the points axis?
    positions = _tile._convert_image_coords_to_positions(microscope, settings, image, [Point(f.px.y, f.px.x) for f in features])

    return positions, features

def _calculate_intersection(masks: list[np.ndarray]) -> np.ndarray:

    intersection = masks[0]
    for mask in masks[1:]:
        intersection = intersection + mask
        intersection[intersection < 2] = 0
        intersection[intersection == 2] = 1
    return intersection


### SEGMENTATION TOOLS

import copy
import glob
import os

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np

from skimage import measure

from fibsem.segmentation import config as segcfg
from fibsem.segmentation.utils import decode_segmap_v2



def plot_instance_masks(image: np.ndarray, mask: np.ndarray, objects: list[dict], ncols:int = 10, show: bool = True):
    """Plot image with instance masks overlayed"""
    # plot instance masks
    n_objects = min(len(objects), ncols) # limit to 10 objects
    fig, ax = plt.subplots(1, n_objects, figsize=(15, 7))
    for i, obj in enumerate(objects):

        if i >= n_objects:
            break

        c = obj["class"]
        instance = obj["instance"]
        imask = np.array(obj["mask"])

        mask = np.zeros_like(mask)
        mask[imask[:, 0], imask[:, 1]] = 1

        tmp_cmap_rgb = [(0, 0, 0), segcfg.CLASS_COLORS_RGB[c]] # black and class color

        # plot image
        if len(objects) == 1:
            axx = ax
        else:
            axx = ax[i]
            
        axx.imshow(image, cmap="gray", alpha=0.7)
        axx.imshow(decode_segmap_v2(mask, tmp_cmap_rgb), alpha=0.3)
        axx.set_title(f"{segcfg.CLASS_LABELS[c]}_{instance}")

        # axes off
        axx.axis("off")

    plt.subplots_adjust(wspace=0.05, hspace=0.05)
    
    if show:
        plt.show()

    return fig


def plot_instance_masks_grid(image: np.ndarray, mask: np.ndarray, 
                             objects: list[dict], ncols: int = 5, 
                             show: bool = True):
    """Plot image with instance masks overlayed"""
    # plot instance masks
    # craete grid from objects list, with
    #  nrows = n unique classes, and 
    # ncols = n unique instance of eaach class
    
    # get unique classes and instances

    # TODO: add class names to title / axes

    plot_objects = {}
    for obj in objects:
        c = obj["class"]
        instance = obj["instance"]
        if c not in plot_objects.keys():
            plot_objects[c] = []
        plot_objects[c].append(obj)

    nr = len(plot_objects.keys())
    nc = min(ncols, max([len(plot_objects[c]) for c in plot_objects.keys()])) # limit to ncols objects
    fig, ax = plt.subplots(nr, nc, figsize=(10, 10))

    for i, c in enumerate(plot_objects.keys()):

        for j, obj in enumerate(plot_objects[c]):

            if j >= nc:
                continue

            c = obj["class"]
            instance = obj["instance"]
            imask = np.array(obj["mask"])

            mask = np.zeros_like(mask)
            mask[imask[:, 0], imask[:, 1]] = 1

            tmp_cmap_rgb = [(0, 0, 0), segcfg.CLASS_COLORS_RGB[c]] # black and class color

            # plot image
            if len(objects) == 1:
                axx = ax
            elif nc == 1:
                axx = ax[i]
            elif nr == 1:
                axx = ax[j]
            else:
                axx = ax[i][j]


            axx.imshow(image, cmap="gray", alpha=0.7)
            axx.imshow(decode_segmap_v2(mask, tmp_cmap_rgb), alpha=0.3)
            # axx.set_title(f"{segcfg.CLASS_LABELS[c]}_{instance}")

            # axes off
            axx.axis("off")

            # set class label has yaxis title
            if j == 0:
                axx.set_ylabel(segcfg.CLASS_LABELS[c], rotation=0, ha="right", va="center", fontsize=20)
                

        # if not enough objects to fill the grid, remove the remaining axes
        if len(plot_objects[c]) < nc:
            for j in range(len(plot_objects[c]), nc):
                ax[i][j].axis("off")

    plt.subplots_adjust(wspace=0.05, hspace=0.05)
    
    if show:
        plt.show()

    return fig

def plot_bounding_boxes(image: np.ndarray, mask: np.ndarray, objects: list[dict], show:bool = True):
    """Plot image with bounding boxes around detected objects"""
    
    # plot bounding boxes
    fig = plt.figure(figsize=(10, 10))
    plt.imshow(image, cmap="gray", alpha=0.7)
    plt.imshow(decode_segmap_v2(mask, segcfg.CLASS_COLORS_RGB), alpha=0.3)            
    

    n_objects = len(objects)
    classes = []

    for obj in objects:
        c = obj["class"]
        instance = obj["instance"]
        (ystart, xstart), (ystop, xstop) = obj["bbox"]

        # if the class is the same as the name, only plot the label once
        if segcfg.CLASS_LABELS[c] in classes:
            label = None
        else: 
            cls_name = segcfg.CLASS_LABELS[c]
            # limit the number of instances to 5
            label = f"{cls_name}_{instance}" if n_objects <= 5 else f"{cls_name}"
            classes.append(cls_name)

        # plot bounding box as rectangle
        rect = patches.Rectangle((xstart, ystart), xstop - xstart, ystop - ystart, 
                                linewidth=1, linestyle="--", 
                                edgecolor=segcfg.CLASS_COLORS[c], 
                                facecolor='none', 
                                label=label)
        plt.gca().add_patch(rect)

    plt.legend(loc="best")
    if show:
        plt.show()

    return fig

def plot_keypoints(mask: np.ndarray, objects: list[dict]):
    for obj in objects:

        c = obj["class"]
        instance = obj["instance"]
        imask = np.array(obj["mask"])
        keypoints = obj["keypoints"]

        mask = np.zeros_like(mask)
        mask[imask[:, 0], imask[:, 1]] = 1   
        
        # filter points for specific classes
        mmap = {
            "lamella": ["centre", "bottom_edge", "top_edge", "left_edge", "right_edge"],
            "manipulator": ["top_left", "top_right", "bottom_left", "bottom_right"],
            "copper_adaptor": ["centre", "bottom_edge", "top_edge"],
        }   

        cname = segcfg.CLASS_LABELS[c]

        if cname not in mmap.keys():
            mmap = {cname: ["centre"]}

        plt.title(f"{cname}_{instance}")
        plt.imshow(mask, cmap="gray") # TODO: decode segmap?
            
        for l, p in keypoints.items():
            
            if l not in mmap[cname]:
                continue
                
            l = l.title().replace("_", "")
            c = cname.title().replace("_", "")
            plt.plot(p[0], p[1], "+", label=f"{c}{l}")
        plt.legend()
        plt.show()

def get_objects(mask: np.ndarray, ignore_classes: list[int] = [0, 3], min_pixels: int = 100):
    """ Extract individual objects from a mask """

    classes = [c for c in np.unique(mask) if c not in ignore_classes] # TODO: make this more granular, e.g. ignore-bbox, ignore-keypoints, etc.

    objects = []
    for c in classes:
        mask_c = mask == c

        # extract instance labels
        label = measure.label(mask_c).astype(np.uint8)
        instances = np.unique(label)  # 0 = background, 1 = instance 1, 2 = instance 2, etc.
        for instance in instances[1:]: # skip background
            
            # get bounding box
            instance_mask = np.argwhere(label==instance)

            # check how many pixels are in the instance mask
            if len(instance_mask) < min_pixels:
                logging.debug(f"skipping instance {instance} of class {c} because it has {len(instance_mask)} pixels. The minimum is {min_pixels}")
                continue

            (ystart, xstart), (ystop, xstop) = instance_mask.min(0), instance_mask.max(0) + 1
            bbox = [(ystart, xstart), (ystop, xstop)]
            
            # get keypoints
            kmask = np.zeros_like(mask)
            kmask[instance_mask[:, 0], instance_mask[:, 1]] = 1   
            keypoints = get_keypoints(kmask) # TODO: fix inefficiency here

            # save instance masks
            odict = {"class": c, "instance": instance, 
                     "mask": instance_mask, 
                     "bbox": bbox, "keypoints": keypoints}
            objects.append(copy.deepcopy(odict))

    return objects

from scipy import ndimage

def get_keypoints(mask: np.ndarray) -> dict:
        """Get keypoints from instance mask.
        Args:
            mask (np.ndarray): instance mask    
        """

        # get center of mass, from list of points
        cy, cx = np.array(ndimage.center_of_mass(mask))

        # get edges
        imask = np.argwhere(mask==1) # probably inefficient to recompute this

        ymin = np.max(imask[:, 0])
        ymax = np.min(imask[:, 0])
        xmin = np.min(imask[:, 1])
        xmax = np.max(imask[:, 1])

        centre = cx, cy
        bot = cx, ymin
        top = cx, ymax
        left = xmin, cy
        right = xmax, cy

        # get width and height
        w = xmax - xmin
        h = ymax - ymin   
        
        # corners
        tl = xmin, ymin
        tr = xmax, ymin
        bl = xmin, ymax
        br = xmax, ymax

        pts = [centre, bot, top, left, right, tl, tr, bl, br]
        labels = ["centre", "bottom_edge", "top_edge", "left_edge", "right_edge", 
                "top_left", "top_right", "bottom_left", "bottom_right"]

        keypoints = dict(zip(labels, pts))

        return keypoints 


# SAVE/ LOAD data as json
import json

def save_json(data, filename):
    # write np array to json
    # https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
    class NumpyEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, np.ndarray):
                return obj.tolist()
            if isinstance(obj, (np.uint8, np.int64, np.uint64)):
                return int(obj)
            return json.JSONEncoder.default(self, obj)
    

    # save json  to file 
    with open(filename, "w") as f:
        json.dump(data, f, cls=NumpyEncoder, indent=4)

def load_json(filename):
    with open(filename, "r") as f:
        data = json.load(f)
    return data


from tqdm import tqdm 

def generate_segmentation_objects(data_path: str, labels_path: str, dataset_json_path: str, min_pixels: int = 100, save: bool=True):
    image_filenames = sorted(glob.glob(os.path.join(data_path, "*.tif")))
    label_filenames = sorted(glob.glob(os.path.join(labels_path, "*.tif")))

    filenames = list(zip(image_filenames, label_filenames)) # TDOO: we dont actually need image files for this, just the labels?
    dat = []

    progress = tqdm(filenames)
    for img_fname, label_fname in progress:
        progress.set_description(f"Processing {os.path.basename(img_fname)}")
        
        image = tff.imread(img_fname)
        mask = tff.imread(label_fname)

        # get objects
        objects = get_objects(mask, min_pixels=min_pixels)

        # save 
        dat.append(copy.deepcopy({"filename": os.path.basename(img_fname), 
                                "path": os.path.dirname(img_fname), 
                                "mask_filename": os.path.basename(label_fname), 
                                "mask_path": os.path.dirname(label_fname),
                                "objects": objects}))

    if save:
        print(f"Saving data.json to {dataset_json_path}")
        save_json(dat, dataset_json_path)

    return dat


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemFeatureDetectionUI.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(429, 648)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.comboBox_features = QtWidgets.QComboBox(self.centralwidget)
        self.comboBox_features.setObjectName("comboBox_features")
        self.gridLayout.addWidget(self.comboBox_features, 3, 0, 1, 2)
        self.label_header_features = QtWidgets.QLabel(self.centralwidget)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_header_features.setFont(font)
        self.label_header_features.setObjectName("label_header_features")
        self.gridLayout.addWidget(self.label_header_features, 0, 0, 1, 2)
        self.pushButton_edit_feature = QtWidgets.QPushButton(self.centralwidget)
        self.pushButton_edit_feature.setObjectName("pushButton_edit_feature")
        self.gridLayout.addWidget(self.pushButton_edit_feature, 2, 1, 1, 1)
        self.label_features_header = QtWidgets.QLabel(self.centralwidget)
        self.label_features_header.setObjectName("label_features_header")
        self.gridLayout.addWidget(self.label_features_header, 1, 0, 1, 2)
        self.pushButton_add_feature = QtWidgets.QPushButton(self.centralwidget)
        self.pushButton_add_feature.setObjectName("pushButton_add_feature")
        self.gridLayout.addWidget(self.pushButton_add_feature, 2, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 7, 0, 1, 2)
        self.label_instructions = QtWidgets.QLabel(self.centralwidget)
        self.label_instructions.setWordWrap(True)
        self.label_instructions.setObjectName("label_instructions")
        self.gridLayout.addWidget(self.label_instructions, 6, 0, 1, 2)
        self.spinBox_data_progress = QtWidgets.QSpinBox(self.centralwidget)
        self.spinBox_data_progress.setAlignment(QtCore.Qt.AlignCenter)
        self.spinBox_data_progress.setMaximum(999999999)
        self.spinBox_data_progress.setObjectName("spinBox_data_progress")
        self.gridLayout.addWidget(self.spinBox_data_progress, 5, 1, 1, 1)
        self.label_data_progress = QtWidgets.QLabel(self.centralwidget)
        self.label_data_progress.setObjectName("label_data_progress")
        self.gridLayout.addWidget(self.label_data_progress, 5, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 429, 22))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionLoad_Image_Directory = QtWidgets.QAction(MainWindow)
        self.actionLoad_Image_Directory.setObjectName("actionLoad_Image_Directory")
        self.actionLoad_CSV = QtWidgets.QAction(MainWindow)
        self.actionLoad_CSV.setObjectName("actionLoad_CSV")
        self.menuFile.addAction(self.actionLoad_Image_Directory)
        self.menuFile.addAction(self.actionLoad_CSV)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_header_features.setText(_translate("MainWindow", "OpenFIBSEM Keypoint Labelling"))
        self.pushButton_edit_feature.setText(_translate("MainWindow", "Edit"))
        self.label_features_header.setText(_translate("MainWindow", "Features"))
        self.pushButton_add_feature.setText(_translate("MainWindow", "Add"))
        self.label_instructions.setText(_translate("MainWindow", "Instructions"))
        self.label_data_progress.setText(_translate("MainWindow", "Current Image (Total: N)"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.actionLoad_Image_Directory.setText(_translate("MainWindow", "Load Image Directory"))
        self.actionLoad_CSV.setText(_translate("MainWindow", "Load CSV"))



import logging
import napari
import napari.utils.notifications
import numpy as np
from PyQt5 import QtWidgets
from fibsem import config as cfg
from fibsem import constants, conversions, utils
from fibsem.microscope import FibsemMicroscope, TescanMicroscope, ThermoMicroscope, DemoMicroscope
from fibsem.structures import MicroscopeSettings
from fibsem.ui import FibsemCryoDepositionWidget_qt
from fibsem import gis


class FibsemCryoDepositionWidget(FibsemCryoDepositionWidget_qt.Ui_Dialog, QtWidgets.QDialog):
    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        settings: MicroscopeSettings = None,
        parent=None,
    ):
        super(FibsemCryoDepositionWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.setWindowTitle("Cryo Deposition")

        self.microscope = microscope
        self.settings = settings

        self.setup_connections()

    def setup_connections(self):

        self.pushButton_run_sputter.clicked.connect(self.run_sputter)

        positions = utils.load_yaml(cfg.POSITION_PATH)
        self.comboBox_sputter_stage_position.addItems([p["name"] for p in positions])


    def _get_protocol_from_ui(self):
        
        protocol = {
            "application_file": self.lineEdit_sputter_application_file.text(),
            "gas": self.lineEdit_sputter_gas.text(),
            "position": self.lineEdit_sputter_gis_position.text(),
            "name": self.comboBox_sputter_stage_position.currentText(),
            "time": self.doubleSpinBox_sputter_time.value(),
            "hfw": self.doubleSpinBox_sputter_hfw.value() * constants.MICRO_TO_SI,
            "beam_current": self.doubleSpinBox_sputter_beam_current.value() * constants.NANO_TO_SI,
            "length": self.doubleSpinBox_sputter_length.value() * constants.MICRO_TO_SI,

        }

        return protocol

    # TODO: thread this
    def run_sputter(self):
        
        protocol = self._get_protocol_from_ui()

        gis.cryo_deposition(self.microscope, protocol, name=protocol["name"])




def main():

    viewer = napari.Viewer(ndisplay=2)
    microscope, settings = utils.setup_session()
    cryo_sputter_widget = FibsemCryoDepositionWidget(microscope, settings)
    viewer.window.add_dock_widget(
        cryo_sputter_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemGISWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(506, 637)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.timeDuration_spinbox = QtWidgets.QDoubleSpinBox(Form)
        self.timeDuration_spinbox.setObjectName("timeDuration_spinbox")
        self.gridLayout_2.addWidget(self.timeDuration_spinbox, 11, 1, 1, 1)
        self.beamtype_combobox = QtWidgets.QComboBox(Form)
        self.beamtype_combobox.setObjectName("beamtype_combobox")
        self.beamtype_combobox.addItem("")
        self.beamtype_combobox.addItem("")
        self.gridLayout_2.addWidget(self.beamtype_combobox, 12, 1, 1, 1)
        self.label_position = QtWidgets.QLabel(Form)
        self.label_position.setObjectName("label_position")
        self.gridLayout_2.addWidget(self.label_position, 2, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 18, 0, 1, 2)
        self.warm_button = QtWidgets.QPushButton(Form)
        self.warm_button.setObjectName("warm_button")
        self.gridLayout_2.addWidget(self.warm_button, 10, 1, 1, 1)
        self.hfw_spinbox = QtWidgets.QDoubleSpinBox(Form)
        self.hfw_spinbox.setMaximum(10000.0)
        self.hfw_spinbox.setObjectName("hfw_spinbox")
        self.gridLayout_2.addWidget(self.hfw_spinbox, 13, 1, 1, 1)
        self.position_combobox = QtWidgets.QComboBox(Form)
        self.position_combobox.setObjectName("position_combobox")
        self.gridLayout_2.addWidget(self.position_combobox, 2, 1, 1, 1)
        self.run_button = QtWidgets.QPushButton(Form)
        self.run_button.setObjectName("run_button")
        self.gridLayout_2.addWidget(self.run_button, 17, 0, 1, 2)
        self.move_GIS_button = QtWidgets.QPushButton(Form)
        self.move_GIS_button.setObjectName("move_GIS_button")
        self.gridLayout_2.addWidget(self.move_GIS_button, 5, 1, 1, 1)
        self.label_title = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout_2.addWidget(self.label_title, 0, 0, 1, 2)
        self.pushButton_settings_update = QtWidgets.QPushButton(Form)
        self.pushButton_settings_update.setObjectName("pushButton_settings_update")
        self.gridLayout_2.addWidget(self.pushButton_settings_update, 16, 0, 1, 2)
        self.label_gas = QtWidgets.QLabel(Form)
        self.label_gas.setObjectName("label_gas")
        self.gridLayout_2.addWidget(self.label_gas, 8, 0, 1, 1)
        self.blankBeamcheckbox = QtWidgets.QCheckBox(Form)
        self.blankBeamcheckbox.setObjectName("blankBeamcheckbox")
        self.gridLayout_2.addWidget(self.blankBeamcheckbox, 14, 1, 1, 1)
        self.label_beam_type = QtWidgets.QLabel(Form)
        self.label_beam_type.setObjectName("label_beam_type")
        self.gridLayout_2.addWidget(self.label_beam_type, 12, 0, 1, 1)
        self.current_position_label = QtWidgets.QLabel(Form)
        self.current_position_label.setObjectName("current_position_label")
        self.gridLayout_2.addWidget(self.current_position_label, 5, 0, 1, 1)
        self.label_hfw = QtWidgets.QLabel(Form)
        self.label_hfw.setObjectName("label_hfw")
        self.gridLayout_2.addWidget(self.label_hfw, 13, 0, 1, 1)
        self.app_file_label = QtWidgets.QLabel(Form)
        self.app_file_label.setObjectName("app_file_label")
        self.gridLayout_2.addWidget(self.app_file_label, 15, 0, 1, 1)
        self.app_file_combobox = QtWidgets.QComboBox(Form)
        self.app_file_combobox.setObjectName("app_file_combobox")
        self.gridLayout_2.addWidget(self.app_file_combobox, 15, 1, 1, 1)
        self.temp_label = QtWidgets.QLabel(Form)
        self.temp_label.setObjectName("temp_label")
        self.gridLayout_2.addWidget(self.temp_label, 10, 0, 1, 1)
        self.gas_combobox = QtWidgets.QComboBox(Form)
        self.gas_combobox.setObjectName("gas_combobox")
        self.gridLayout_2.addWidget(self.gas_combobox, 8, 1, 1, 1)
        self.label_time = QtWidgets.QLabel(Form)
        self.label_time.setObjectName("label_time")
        self.gridLayout_2.addWidget(self.label_time, 11, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.beamtype_combobox.setItemText(0, _translate("Form", "ELECTRON"))
        self.beamtype_combobox.setItemText(1, _translate("Form", "ION"))
        self.label_position.setText(_translate("Form", "Position"))
        self.warm_button.setText(_translate("Form", "Warm up"))
        self.run_button.setText(_translate("Form", "Run"))
        self.move_GIS_button.setText(_translate("Form", "Move to Position"))
        self.label_title.setText(_translate("Form", "Gas Injection System"))
        self.pushButton_settings_update.setText(_translate("Form", "Update Settings"))
        self.label_gas.setText(_translate("Form", "Select Gas"))
        self.blankBeamcheckbox.setText(_translate("Form", "Blank Beam"))
        self.label_beam_type.setText(_translate("Form", "BeamType"))
        self.current_position_label.setText(_translate("Form", "Current Position:"))
        self.label_hfw.setText(_translate("Form", "HFW (um)"))
        self.app_file_label.setText(_translate("Form", "App. File"))
        self.temp_label.setText(_translate("Form", "Temp: "))
        self.label_time.setText(_translate("Form", "Time Duration (s)"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'CurrentAlignmentWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_BeamAlignment(object):
    def setupUi(self, BeamAlignment):
        BeamAlignment.setObjectName("BeamAlignment")
        BeamAlignment.resize(398, 380)
        self.gridLayout = QtWidgets.QGridLayout(BeamAlignment)
        self.gridLayout.setObjectName("gridLayout")
        self.label_3 = QtWidgets.QLabel(BeamAlignment)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 4, 0, 1, 1)
        self.pushButton_align_beam = QtWidgets.QPushButton(BeamAlignment)
        self.pushButton_align_beam.setObjectName("pushButton_align_beam")
        self.gridLayout.addWidget(self.pushButton_align_beam, 6, 0, 1, 2)
        self.checkBox_overlay = QtWidgets.QCheckBox(BeamAlignment)
        self.checkBox_overlay.setObjectName("checkBox_overlay")
        self.gridLayout.addWidget(self.checkBox_overlay, 2, 1, 1, 1)
        self.listdone = QtWidgets.QPlainTextEdit(BeamAlignment)
        self.listdone.setObjectName("listdone")
        self.gridLayout.addWidget(self.listdone, 7, 0, 1, 2)
        self.label = QtWidgets.QLabel(BeamAlignment)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.comboBox_aligned_current = QtWidgets.QComboBox(BeamAlignment)
        self.comboBox_aligned_current.setObjectName("comboBox_aligned_current")
        self.gridLayout.addWidget(self.comboBox_aligned_current, 4, 1, 1, 1)
        self.comboBox_ref_current = QtWidgets.QComboBox(BeamAlignment)
        self.comboBox_ref_current.setObjectName("comboBox_ref_current")
        self.gridLayout.addWidget(self.comboBox_ref_current, 3, 1, 1, 1)
        self.pushButton_take_images = QtWidgets.QPushButton(BeamAlignment)
        self.pushButton_take_images.setObjectName("pushButton_take_images")
        self.gridLayout.addWidget(self.pushButton_take_images, 5, 0, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 0, 1, 2)
        self.label_2 = QtWidgets.QLabel(BeamAlignment)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 3, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 8, 0, 1, 2)

        self.retranslateUi(BeamAlignment)
        QtCore.QMetaObject.connectSlotsByName(BeamAlignment)
        BeamAlignment.setTabOrder(self.checkBox_overlay, self.comboBox_ref_current)
        BeamAlignment.setTabOrder(self.comboBox_ref_current, self.comboBox_aligned_current)
        BeamAlignment.setTabOrder(self.comboBox_aligned_current, self.pushButton_take_images)
        BeamAlignment.setTabOrder(self.pushButton_take_images, self.pushButton_align_beam)
        BeamAlignment.setTabOrder(self.pushButton_align_beam, self.listdone)

    def retranslateUi(self, BeamAlignment):
        _translate = QtCore.QCoreApplication.translate
        BeamAlignment.setWindowTitle(_translate("BeamAlignment", "Beam alignment"))
        self.label_3.setText(_translate("BeamAlignment", "Aligned Current (Right):"))
        self.pushButton_align_beam.setText(_translate("BeamAlignment", "Align Beam"))
        self.checkBox_overlay.setText(_translate("BeamAlignment", "Overlay Images"))
        self.label.setText(_translate("BeamAlignment", "FIB Current alignment"))
        self.pushButton_take_images.setText(_translate("BeamAlignment", "Take Images"))
        self.label_2.setText(_translate("BeamAlignment", "Reference Current (Left):"))



import io
import logging
import os
import time
import traceback
from copy import deepcopy
from pathlib import Path

import napari
import napari.utils.notifications
from napari.qt.threading import thread_worker
import numpy as np
import yaml
from PIL import Image
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtGui import QImage, QPixmap
from fibsem import config as cfg
from fibsem import constants, conversions
from fibsem.microscope import FibsemMicroscope
from fibsem.structures import (BeamType, FibsemStagePosition,
                               MicroscopeSettings, MovementMode, Point, FibsemImage)
from fibsem.ui.FibsemImageSettingsWidget import FibsemImageSettingsWidget
from fibsem.ui.qtdesigner_files import FibsemMovementWidget
from fibsem.ui.utils import _get_file_ui, _get_save_file_ui
from fibsem.imaging._tile import _plot_positions, _minimap 
from fibsem.ui import utils as ui_utils 
from fibsem.ui.utils import message_box_ui
from matplotlib.backends.backend_agg import FigureCanvasAgg
from fibsem.ui import _stylesheets
import fibsem.utils as utils

def log_status_message(step: str):
    logging.debug(
        f"STATUS | Movement Widget | {step}"
    )


class FibsemMovementWidget(FibsemMovementWidget.Ui_Form, QtWidgets.QWidget):
    move_signal = QtCore.pyqtSignal()
    movement_notification_signal = QtCore.pyqtSignal(str)
    positions_signal = QtCore.pyqtSignal(object)

    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        settings: MicroscopeSettings = None,
        viewer: napari.Viewer = None,
        image_widget: FibsemImageSettingsWidget = None, 
        parent=None,
    ):
        super(FibsemMovementWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.parent = parent
        self.microscope = microscope
        self.settings = settings
        self.viewer = viewer
        self.image_widget = image_widget

        self.setup_connections()
        self.image_widget.picture_signal.connect(self.update_ui)
        self.positions = []
        self.minimap_image = None
        autoload_positions: bool = True # TODO: enable this
        if autoload_positions:
            self.import_positions(cfg.POSITION_PATH)
        self.update_ui()

    def setup_connections(self):

        # buttons
        self.pushButton_move.clicked.connect(self.move_to_position)
        self.pushButton_move.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.pushButton_move_flat_ion.clicked.connect(self.move_flat_to_beam)
        self.pushButton_move_flat_ion.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.pushButton_move_flat_electron.clicked.connect(self.move_flat_to_beam)
        self.pushButton_move_flat_electron.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        
        # minimap
        self.pushButton_load_image_minimap.clicked.connect(self.load_image)
        self.pushButton_load_image_minimap.setVisible(cfg._MINIMAP_VISUALISATION) # DISABLE_MINIMAP IN UI
        self.label_4.setVisible(cfg._MINIMAP_VISUALISATION) # DISABLE_MINIMAP IN UI
        self.label_minimap.setVisible(cfg._MINIMAP_VISUALISATION) # DISABLE_MINIMAP IN UI

        self.pushButton_refresh_stage_position_data.clicked.connect(self.update_ui)
        self.pushButton_refresh_stage_position_data.setStyleSheet(_stylesheets._GRAY_PUSHBUTTON_STYLE)

        # register mouse callbacks
        self.image_widget.eb_layer.mouse_double_click_callbacks.append(self._double_click)
        self.image_widget.ib_layer.mouse_double_click_callbacks.append(self._double_click)

        # disable ui elements
        self.label_movement_instructions.setText("Double click to move. Alt + Double Click in the Ion Beam to Move Vertically")

        # positions
        self.comboBox_positions.currentIndexChanged.connect(self.select_position)
        self.pushButton_save_position.clicked.connect(lambda: self.add_position(position=None))
        self.pushButton_save_position.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.pushButton_remove_position.clicked.connect(self.delete_position)
        self.pushButton_remove_position.setStyleSheet(_stylesheets._RED_PUSHBUTTON_STYLE)
        self.pushButton_go_to.clicked.connect(lambda: self.go_to_saved_position(None))
        self.pushButton_go_to.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.pushButton_export.clicked.connect(self.export_positions)
        self.pushButton_export.setStyleSheet(_stylesheets._GRAY_PUSHBUTTON_STYLE)
        self.pushButton_import.clicked.connect(self.import_positions)
        self.pushButton_import.setStyleSheet(_stylesheets._GRAY_PUSHBUTTON_STYLE)
        self.pushButton_update_position.clicked.connect(self.update_saved_position)
        self.pushButton_update_position.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)

        self.movement_notification_signal.connect(self.update_moving_ui)

    def _toggle_interactions(self, enable: bool, caller: str = None):
        
        self.pushButton_move.setEnabled(enable)
        self.pushButton_move_flat_ion.setEnabled(enable)
        self.pushButton_move_flat_electron.setEnabled(enable)
        self.pushButton_go_to.setEnabled(enable)
        if caller is None:
            self.parent.milling_widget._toggle_interactions(enable, caller="movement")
            self.parent.image_widget._toggle_interactions(enable, caller="movement")
        if enable:
            self.pushButton_move.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.pushButton_move_flat_ion.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
            self.pushButton_move_flat_electron.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
            self.pushButton_go_to.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        else:
            self.pushButton_move.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_move_flat_ion.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_move_flat_electron.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_go_to.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)

    def move_to_position(self):
        worker = self.move_worker()
        worker.finished.connect(self.run_moving_finished)
        # worker.yielded.connect(self.update_moving_ui)
        worker.start()
    
    @thread_worker
    def move_worker(self):
        self._toggle_interactions(False)
        stage_position = self.get_position_from_ui()
        self.movement_notification_signal.emit(f"Moving to {stage_position}")
        self.microscope.move_stage_absolute(stage_position)
        log_status_message(f"MOVED_TO_{stage_position}")
        self.movement_notification_signal.emit("Move finished, taking new images")
        self.update_ui_after_movement()

    def run_moving_finished(self):
        self.positions_signal.emit(None)
        if self.parent.image_widget.TAKING_IMAGES:
            return
        self._toggle_interactions(True)
        if self.parent.image_widget._LIVE_IMAGING:
            self.parent.image_widget.live_imaging()


    def update_moving_ui(self, msg: str):
        logging.info(msg)
        napari.utils.notifications.notification_manager.records.clear()
        napari.utils.notifications.show_info(msg)

    def update_ui(self):

        stage_position: FibsemStagePosition = self.microscope.get_stage_position()

        self.doubleSpinBox_movement_stage_x.setValue(stage_position.x * constants.SI_TO_MILLI)
        self.doubleSpinBox_movement_stage_y.setValue(stage_position.y * constants.SI_TO_MILLI)
        self.doubleSpinBox_movement_stage_z.setValue(stage_position.z * constants.SI_TO_MILLI)
        self.doubleSpinBox_movement_stage_rotation.setValue(np.rad2deg(stage_position.r))
        self.doubleSpinBox_movement_stage_tilt.setValue(np.rad2deg(stage_position.t))

        # NOTE (pc): temporary to reduce number of updates    
        if self.sender() is None:
            self.minimap()

        _SAVED_POSITIONS = bool(len(self.positions))
        # self.pushButton_save_position.setVisible(_SAVED_POSITIONS) # always visible
        self.pushButton_remove_position.setEnabled(_SAVED_POSITIONS)
        self.pushButton_go_to.setVisible(_SAVED_POSITIONS)
        self.pushButton_update_position.setVisible(_SAVED_POSITIONS)
        self.pushButton_export.setVisible(_SAVED_POSITIONS)
        self.comboBox_positions.setVisible(_SAVED_POSITIONS)
        self.pushButton_import.setVisible(_SAVED_POSITIONS)
        self.label_saved_positions.setVisible(_SAVED_POSITIONS)
        self.label_current_position.setVisible(_SAVED_POSITIONS)
        self.lineEdit_position_name.setVisible(_SAVED_POSITIONS)
    
    def get_position_from_ui(self):

        stage_position = FibsemStagePosition(
            x=self.doubleSpinBox_movement_stage_x.value() * constants.MILLI_TO_SI,
            y=self.doubleSpinBox_movement_stage_y.value() * constants.MILLI_TO_SI,
            z=self.doubleSpinBox_movement_stage_z.value() * constants.MILLI_TO_SI,
            r=np.deg2rad(self.doubleSpinBox_movement_stage_rotation.value()),
            t=np.deg2rad(self.doubleSpinBox_movement_stage_tilt.value()),
            coordinate_system="RAW",

        )

        return stage_position

    def _double_click(self, layer, event):
        worker = self._double_click_worker(layer, event)
        worker.finished.connect(self.run_moving_finished)
        # worker.yielded.connect(self.update_moving_ui)
        if self.parent.image_widget._LIVE_IMAGING:
            self.parent.image_widget.stop_event.set()
        worker.start()

    @thread_worker
    def _double_click_worker(self, layer, event):
        
        if event.button != 1 or "Shift" in event.modifiers:
            return
        self._toggle_interactions(False)
        # get coords
        coords = layer.world_to_data(event.position)

        # TODO: dimensions are mixed which makes this confusing to interpret... resolve
        coords, beam_type, image = self.image_widget.get_data_from_coord(coords)
        self.movement_notification_signal.emit("Click to move in progress")
        if beam_type is None:
            napari.utils.notifications.show_info(
                f"Clicked outside image dimensions. Please click inside the image to move."
            )
            return

        point = conversions.image_to_microscope_image_coordinates(
            Point(x=coords[1], y=coords[0]), image.data, image.metadata.pixel_size.x,
        )

        # move
        self.movement_mode = MovementMode.Vertical if "Alt" in event.modifiers else MovementMode.Stable

        logging.debug({
            "msg": "stage_movement",                    # message type
            "movement_mode": self.movement_mode.name,   # movement mode
            "beam_type": beam_type.name,                # beam type
            "dm": point.to_dict(),                      # shift in microscope coordinates
            "coords": {"x": coords[1], "y": coords[0]}, # coords in image coordinates
        })
        
        log_status_message(f"MOVING_{self.movement_mode.name}_BY_{point.x:.2e}, {point.y:.2e} | {beam_type}")
        self.movement_notification_signal.emit("Moving stage ")
        # eucentric is only supported for ION beam
        if beam_type is BeamType.ION and self.movement_mode is MovementMode.Vertical:
            self.microscope.vertical_move(dx=point.x, dy=-point.y
            )

        else:
            # corrected stage movement
            self.microscope.stable_move(
                dx=point.x,
                dy=point.y,
                beam_type=beam_type,
            )
        self.movement_notification_signal.emit("Move finished, updating UI")
        self.update_ui_after_movement()

    def select_position(self):
        if self.comboBox_positions.currentIndex() != -1:
            position = self.positions[self.comboBox_positions.currentIndex()]
            self.label_current_position.setText(f"x={position.x*constants.METRE_TO_MILLIMETRE:.3f}, y={position.y*constants.METRE_TO_MILLIMETRE:.3f}, z={position.z*constants.METRE_TO_MILLIMETRE:.3f}, r={position.r*constants.RADIANS_TO_DEGREES:.1f}, t={position.t*constants.RADIANS_TO_DEGREES:.1f}")
            self.lineEdit_position_name.setText(position.name)

    def add_position(self, position: FibsemStagePosition = None):

        if not isinstance(position, FibsemStagePosition):
            position = self.microscope.get_stage_position()
            name = f"Position {len(self.positions):02d}"
            # if name == "":
                # napari.utils.notifications.show_warning("Please enter a name for the position")
                # return
            position.name = name
        self.positions.append(deepcopy(position))
        self.comboBox_positions.addItem(position.name)
        self.comboBox_positions.setCurrentIndex(self.comboBox_positions.count() - 1)
        self.lineEdit_position_name.setText(position.name)
        logging.info(f"Added position {position.name}")
        self.positions_signal.emit(self.positions)
        self.minimap()
        self.update_ui()

    def delete_position(self):
        del self.positions[self.comboBox_positions.currentIndex()]
        name = self.comboBox_positions.currentText()
        self.comboBox_positions.removeItem(self.comboBox_positions.currentIndex())
        logging.info(f"Removed position {name}")
        self.positions_signal.emit(self.positions)
        self.minimap()
        self.update_ui()

    def update_saved_position(self):
        position = self.microscope.get_stage_position()
        position.name = self.lineEdit_position_name.text()
        if position.name == "":
            napari.utils.notifications.show_warning("Please enter a name for the position")
            return
        self.positions[self.comboBox_positions.currentIndex()] = deepcopy(position)

        # update combobox
        self.comboBox_positions.setItemText(self.comboBox_positions.currentIndex(), position.name)

        self.select_position()
        logging.info(f"Updated position {self.comboBox_positions.currentText()}")
        self.positions_signal.emit(self.positions)
        self.minimap()

    def go_to_saved_position(self, pos:FibsemStagePosition = None):
        worker = self.go_to_saved_position_worker(pos)
        worker.finished.connect(self.run_moving_finished)
        # worker.yielded.connect(self.update_moving_ui)
        worker.start()
    
    @thread_worker
    def go_to_saved_position_worker(self, pos: FibsemStagePosition = None):

        if pos is None:
            pos = self.positions[self.comboBox_positions.currentIndex()]
        self._toggle_interactions(False)
        self.movement_notification_signal.emit(f"Moving to saved position {pos}")
        self.microscope.safe_absolute_stage_movement(pos)
        logging.info(f"Moved to position {pos}")
        self.update_ui_after_movement()


    def export_positions(self):

        path = _get_save_file_ui(msg="Select or create file", 
            path=cfg.POSITION_PATH, 
            _filter="YAML Files (*.yaml)")
        
        if path == '':
            napari.utils.notifications.show_info("No file selected, positions not saved")
            return
        
        response = message_box_ui(text="Do you want to overwrite the file ? Click no to append the new positions to the existing file.", 
            title="Overwrite ?", buttons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        
        # save positions
        utils.save_positions(self.positions, path, overwrite=response)

        logging.info(f"Positions saved to {path}")


    def minimap_window_positions(self,positions):

        self.positions = positions
        self.comboBox_positions.clear()
        for position in positions:
            self.comboBox_positions.addItem(position.name)
        self.minimap()

    def import_positions(self, path: str = None):
        if not isinstance(path, str):
            path = _get_file_ui(msg="Select or create file")
        if path == '':
            napari.utils.notifications.show_info("No file selected, positions not loaded")
            return
        with open(path, "r") as f:
            ddict = yaml.safe_load(f)
        
        for pdict in ddict:
            position = FibsemStagePosition.from_dict(pdict)
            self.positions.append(position)
            self.comboBox_positions.addItem(position.name)
        self.minimap()
        self.positions_signal.emit(self.positions)

    def load_image(self):

        path = ui_utils._get_file_ui( msg="Select image to load", path=cfg.DATA_TILE_PATH, _filter="Image Files (*.tif *.tiff)", parent=self)

        if path == "":
            napari.utils.notifications.show_info(f"No file selected..")
            return

        
        image = FibsemImage.load(path)
        if image.metadata is None:
            napari.utils.notifications.show_error(f"Could not load image {path}. Make sure it is an OpenFibsem Image.")
            return

        self.minimap_image = image
        self.minimap()

    def minimap(self):

        if self.minimap_image is None:
            return
        
        current_position = self.microscope.get_stage_position()
        current_position.name = "Current Position"
        positions = deepcopy(self.positions)
        positions.insert(0, current_position)
        
        qpixmap = _minimap(self.minimap_image, positions)

        self.label_minimap.setPixmap(qpixmap)


    def update_ui_after_movement(self,retake:bool = True): # TODO: PPP Refactor
        # disable taking images after movement here
        if retake is False:
            self.update_ui()
            return
        if self.checkBox_movement_acquire_electron.isChecked() and self.checkBox_movement_acquire_ion.isChecked():
            self.image_widget.take_reference_images()
            return
        if self.checkBox_movement_acquire_electron.isChecked():
            self.image_widget.take_image(BeamType.ELECTRON)
        elif self.checkBox_movement_acquire_ion.isChecked():
            self.image_widget.take_image(BeamType.ION)
        else: 
            self.update_ui()
    
    def _stage_position_moved(self, pos: FibsemStagePosition):
        self.update_ui_after_movement(retake=False)

    def move_flat_to_beam(self):
        beam_type = BeamType.ION if self.sender() == self.pushButton_move_flat_ion else BeamType.ELECTRON
        worker = self.move_flat_to_beam_worker(beam_type)
        worker.finished.connect(self.run_moving_finished)
        # worker.yielded.connect(self.update_moving_ui)
        worker.start()

    @thread_worker
    def move_flat_to_beam_worker(self, beam_type):
        self._toggle_interactions(False)
        self.movement_notification_signal.emit(f"Moving flat to {beam_type.name} beam")
        self.microscope.move_flat_to_beam(beam_type=beam_type)
        self.update_ui_after_movement()



def main():

    viewer = napari.Viewer(ndisplay=2)
    movement_widget = FibsemMovementWidget()
    viewer.window.add_dock_widget(
        movement_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()



# 
import argparse
import os

from fibsem.detection.detection import generate_segmentation_objects

def main():

    parser = argparse.ArgumentParser("Generate segmentation objects from segmentation labels")
    parser.add_argument("--data_path", type=str, help="Path to data directory")
    parser.add_argument("--labels_path", type=str, help="Path to labels directory")
    parser.add_argument("--dataset_json_path", type=str, default=None, help="Path to save dataset json")
    parser.add_argument("--min_pixels", type=int, default=100, help="Minimum number of pixels for an object to be considered")

    args = parser.parse_args()

    if args.dataset_json_path is None:
        args.dataset_json_path = os.path.join(args.data_path, "data.json")
    
    generate_segmentation_objects(
        data_path=args.data_path, 
        labels_path=args.labels_path, 
        dataset_json_path=args.dataset_json_path, 
        min_pixels=args.min_pixels,
        save=True
    )


if __name__ == "__main__":
    main()




import logging
import os
from dataclasses import dataclass
from pathlib import Path
from pprint import pprint

import numpy as np
from fibsem import acquire, alignment, milling, patterning, utils
from fibsem.structures import BeamType, MicroscopeState,  FibsemImage, FibsemStagePosition


@dataclass
class Lamella:
    state: MicroscopeState
    reference_image: FibsemImage
    path: Path

def main():

    PROTOCOL_PATH = os.path.join(os.path.dirname(__file__), "protocol_autolamella.yaml")
    microscope, settings = utils.setup_session(protocol_path=PROTOCOL_PATH)
    
    # move to the milling angle
    stage_position = FibsemStagePosition(
        r=np.deg2rad(settings.protocol["stage_rotation"]),
        t=np.deg2rad(settings.protocol["stage_tilt"])
    )
    microscope.move_stage_absolute(stage_position) # do need a safe version?

    # take a reference image    
    settings.image.filename = "grid_reference"
    settings.image.beam_type = BeamType.ION
    settings.image.hfw = 900e-6
    settings.image.save = True
    acquire.take_reference_images(microscope, settings.image)

    # select positions
    experiment: list[Lamella] = []
    lamella_no = 1
    settings.image.hfw = 80e-6
    base_path = settings.image.path

    while True:
        response = input(f"""Move to the desired position. 
        Do you want to select another lamella? [y]/n {len(experiment)} selected so far.""")

        # store lamella information
        if response.lower() in ["", "y", "yes"]:
            
            # set filepaths
            path = os.path.join(base_path, f"{lamella_no:02d}")
            settings.image.path = path
            settings.image.filename = f"ref_lamella"
            acquire.take_reference_images(microscope, settings.image)

            lamella = Lamella(
                state=microscope.get_microscope_state(),
                reference_image=acquire.new_image(microscope, settings.image),
                path = path
            )
            experiment.append(lamella)
            lamella_no += 1
        else:
            break

    # sanity check
    if len(experiment) == 0:
        logging.info(f"No lamella positions selected. Exiting.")
        return

    # mill (rough, thin, polish)
    workflow_stages = ["rough", "thin", "polish"]
    for stage_no, stage_name in enumerate(workflow_stages):
        
        logging.info(f"Starting milling stage {stage_no}")

        lamella: Lamella
        for lamella_no, lamella in enumerate(experiment):

            logging.info(f"Starting lamella {lamella_no:02d}")

            # return to lamella
            microscope.set_microscope_state(lamella.state)

            # realign
            alignment.beam_shift_alignment(microscope, settings.image, lamella.reference_image)
                       
            if stage_no == 0:
                microexpansion_stage = patterning.get_milling_stages("microexpansion", settings.protocol)
                milling.mill_stage(microscope, microexpansion_stage[0])

            # get trench milling pattern, and mill
            trench_stage = patterning.get_milling_stages("lamella", settings.protocol)[stage_no]
            milling.mill_stage(microscope, trench_stage)

            # retake reference image
            settings.image.path = lamella.path
            settings.image.filename = f"ref_mill_stage_{stage_no:02d}"
            lamella.reference_image = acquire.new_image(microscope, settings.image)

            if stage_no == 3:
                # take final reference images
                settings.image.filename = f"ref_final"
                acquire.take_reference_images(microscope, settings.image)
   
    logging.info(f"Finished autolamella: {settings.protocol['name']}")


if __name__ == "__main__":
    main()




import logging
import time
from PIL import Image
import numpy as np
from copy import deepcopy
import napari
import napari.utils.notifications
from PyQt5 import QtWidgets, QtCore

from fibsem import config as cfg
from fibsem import constants, conversions, milling, patterning, utils
from fibsem.microscope import (DemoMicroscope, FibsemMicroscope,
                               TescanMicroscope, ThermoMicroscope)
from fibsem.patterning import FibsemMillingStage
from fibsem.structures import (BeamType, FibsemMillingSettings,
                               MicroscopeSettings,
                               Point, CrossSectionPattern)
from fibsem.ui.FibsemImageSettingsWidget import FibsemImageSettingsWidget
from fibsem.ui.qtdesigner_files import FibsemMillingWidget
from fibsem.ui.utils import (_draw_patterns_in_napari, _remove_all_layers, 
                            convert_pattern_to_napari_circle, convert_pattern_to_napari_rect, 
                            validate_pattern_placement,
                            _get_directory_ui,_get_file_ui, 
                            _calculate_fiducial_area_v2)
from napari.qt.threading import thread_worker
from fibsem.ui import _stylesheets

_UNSCALED_VALUES  = ["rotation", "size_ratio", "scan_direction", "cleaning_cross_section", 
                     "number", "passes", "n_rectangles", "overlap", "inverted",
                     "n_columns", "n_rows", "cross_section" ]
_ANGLE_KEYS = ["rotation"]
_LINE_KEYS = ["start_x", "start_y", "end_x", "end_y"]

_MILLING_WIDGET_INSTRUCTIONS = """Controls:
Shift + Left Click to Move Selected Pattern
Ctrl + Shift + Left Click to Move All Patterns
Press Run Milling to Start Milling"""

def _scale_value(key, value, scale):
    if key not in _UNSCALED_VALUES:
        return value * scale    
    return value

def log_status_message(stage: FibsemMillingStage, step: str):
    logging.debug(
        f"STATUS | Milling Widget | {stage.name} | {step}"
    )

class FibsemMillingWidget(FibsemMillingWidget.Ui_Form, QtWidgets.QWidget):
    milling_position_changed = QtCore.pyqtSignal()
    _milling_finished = QtCore.pyqtSignal()
    milling_notification = QtCore.pyqtSignal(str)
    _progress_bar_update = QtCore.pyqtSignal(object)
    _progress_bar_start = QtCore.pyqtSignal(object)
    _progress_bar_quit = QtCore.pyqtSignal()

    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        settings: MicroscopeSettings = None,
        viewer: napari.Viewer = None,
        image_widget: FibsemImageSettingsWidget = None,
        protocol: dict = None,
        milling_stages: list[FibsemMillingStage] = [], 
        parent=None,
    ):
        super(FibsemMillingWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.parent = parent
        self.microscope = microscope
        self.settings = settings
        self.viewer = viewer
        self.image_widget = image_widget

        if protocol is None:
            protocol = utils.load_yaml(cfg.PROTOCOL_PATH)
        self.protocol = protocol
        
        self.milling_stages = milling_stages

        self.setup_connections()

        self.update_pattern_ui()

        self.good_copy_pattern = None

        self._UPDATING_PATTERN:bool = False
        self._PATTERN_IS_MOVEABLE: bool = True

    def setup_connections(self):

        self.image_widget.viewer_update_signal.connect(self.update_ui) # this happens after every time the viewer is updated

        # milling
        self.AVAILABLE_MILLING_CURRENTS = self.microscope.get_available_values("current", BeamType.ION)
        # self.comboBox_milling_current.addItems([f"{(current * constants.SI_TO_NANO):.2f}"
        #                                                 for current in self.AVAILABLE_MILLING_CURRENTS])
        # TODO: set limits on milling current
        min_current = self.AVAILABLE_MILLING_CURRENTS[0] * constants.SI_TO_NANO
        max_current = self.AVAILABLE_MILLING_CURRENTS[-1] * constants.SI_TO_NANO
        self.doubleSpinBox_milling_current.setRange(min_current, max_current)
        self.doubleSpinBox_milling_current.setDecimals(4)


        _THERMO = isinstance(self.microscope, ThermoMicroscope)
        _TESCAN = isinstance(self.microscope, TescanMicroscope)

        if isinstance(self.microscope, DemoMicroscope):
            _THERMO, _TESCAN = True, False
        
        # THERMO 
        self.label_application_file.setVisible(_THERMO)
        self.comboBox_application_file.setVisible(_THERMO)
        available_application_files = self.microscope.get_available_values("application_file")
        self.comboBox_application_file.addItems(available_application_files)
        self.comboBox_preset.setVisible(_THERMO)
        self.label_preset.setVisible(_THERMO)
        self.doubleSpinBox_milling_current.setVisible(_THERMO)
        self.label_milling_current.setVisible(_THERMO)
        self.label_voltage.setVisible(_THERMO)
        self.spinBox_voltage.setVisible(_THERMO) # TODO: set this to the available voltages
        self.comboBox_application_file.currentIndexChanged.connect(self.update_settings)
        self.doubleSpinBox_milling_current.valueChanged.connect(self.update_settings)
        self.doubleSpinBox_hfw.valueChanged.connect(self.update_settings)
        if self.comboBox_application_file.findText(self.protocol["milling"]["application_file"]) == -1:
                napari.utils.notifications.show_warning("Application file not available, setting to Si instead")
                self.protocol["milling"]["application_file"] = "Si"
        self.comboBox_application_file.setCurrentText(self.protocol["milling"]["application_file"])
        
        # TESCAN
        self.label_rate.setVisible(_TESCAN)
        self.label_spot_size.setVisible(_TESCAN)
        self.label_dwell_time.setVisible(_TESCAN)
        self.doubleSpinBox_rate.setVisible(_TESCAN)
        self.doubleSpinBox_spot_size.setVisible(_TESCAN)
        self.doubleSpinBox_dwell_time.setVisible(_TESCAN)   
        self.comboBox_preset.setVisible(_TESCAN)
        self.label_preset.setVisible(_TESCAN)
        self.label_spacing.setVisible(_TESCAN)
        self.doubleSpinBox_spacing.setVisible(_TESCAN)
        available_presets = self.microscope.get_available_values("presets")
        self.comboBox_preset.addItems(available_presets)   
        self.doubleSpinBox_rate.valueChanged.connect(self.update_settings)
        self.doubleSpinBox_spot_size.valueChanged.connect(self.update_settings)
        self.doubleSpinBox_dwell_time.valueChanged.connect(self.update_settings)  
        self.comboBox_preset.currentIndexChanged.connect(self.update_settings)
        self.doubleSpinBox_spacing.valueChanged.connect(self.update_settings)
        self.spinBox_voltage.valueChanged.connect(self.update_settings)
        

        # register mouse callbacks
        self.image_widget.eb_layer.mouse_drag_callbacks.append(self._single_click)
        self.image_widget.ib_layer.mouse_drag_callbacks.append(self._single_click)

        #import/export milling stages # TODO: reimplement as protocol export
        # self.pushButton_exportMilling.clicked.connect(self.export_milling_stages)
        # self.pushButton_importMilling.clicked.connect(self.import_milling_stages)

        # new patterns
        self.comboBox_patterns.addItems([pattern.name for pattern in patterning.__PATTERNS__])
        if _TESCAN and not _THERMO:
            index = self.comboBox_patterns.findText("BitmapPattern")
            self.comboBox_patterns.removeItem(index)
        self.comboBox_patterns.currentIndexChanged.connect(lambda: self.update_pattern_ui(None))
    
        # milling stages
        self.pushButton_add_milling_stage.clicked.connect(self.add_milling_stage)
        self.pushButton_add_milling_stage.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.pushButton_remove_milling_stage.clicked.connect(self.remove_milling_stage)
        self.pushButton_remove_milling_stage.setStyleSheet(_stylesheets._RED_PUSHBUTTON_STYLE)
        
        # update ui
        self.pushButton_update_pattern.clicked.connect(lambda: self.update_ui())
        self.pushButton_update_pattern.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.milling_notification.connect(self.update_milling_ui)

        # run milling
        self.pushButton_run_milling.clicked.connect(self.run_milling)
        self.pushButton_run_milling.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)

        if self.milling_stages:
            self.comboBox_milling_stage.addItems([stage.name for stage in self.milling_stages])
            self.update_milling_stage_ui()
        self.comboBox_milling_stage.currentIndexChanged.connect(lambda: self.update_milling_stage_ui())

        # progress bar
        self.progressBar_milling.setVisible(False)
        self._progress_bar_update.connect(self.update_progress_bar)
        self._progress_bar_start.connect(self.start_progress_thread)
        self._progress_bar_quit.connect(self._quit_progress_bar)

        # last
        self.doubleSpinBox_centre_x.setKeyboardTracking(False)
        self.doubleSpinBox_centre_y.setKeyboardTracking(False)
        self.doubleSpinBox_centre_x.valueChanged.connect(self.update_ui_pattern)
        self.doubleSpinBox_centre_y.valueChanged.connect(self.update_ui_pattern)
        self.checkBox_show_milling_crosshair.stateChanged.connect(self.update_ui_pattern)
        self.checkBox_live_update.setChecked(True)
        self.checkBox_live_update.stateChanged.connect(self.toggle_live_update)
        self.toggle_live_update()


        self._available_scan_directions = self.microscope.get_available_values(key="scan_direction")
            
    
        self.label_milling_instructions.setText(_MILLING_WIDGET_INSTRUCTIONS)

    def update_settings(self):
        settings = self.get_milling_settings_from_ui()
        index = self.comboBox_milling_stage.currentIndex()
        if index != -1:
            self.milling_stages[index].milling = settings

    def add_milling_stage(self):
        logging.info("Adding milling stage")

        num = len(self.milling_stages) + 1
        name = f"Milling Stage {num}"
        pattern = patterning.RectanglePattern()
        pattern.define(self.protocol["patterns"]["Rectangle"], Point(0,0))
        milling_stage = FibsemMillingStage(name=name, num=num, pattern=pattern)
        self.milling_stages.append(milling_stage)
        self.comboBox_milling_stage.addItem(name)
        self.comboBox_milling_stage.setCurrentText(name)
        napari.utils.notifications.show_info(f"Added {name}.")
        log_status_message(self.milling_stages[-1], "CREATED_STAGE")

    def remove_milling_stage(self):
        logging.info("Removing milling stage")

        current_index = self.comboBox_milling_stage.currentIndex()
        log_status_message(self.milling_stages[current_index], "REMOVED_STAGE")
        self.milling_stages.pop(current_index)
        self.comboBox_milling_stage.removeItem(current_index)
        napari.utils.notifications.show_info(f"Removed milling stage.")     


    def _remove_all_stages(self):

        self.milling_stages = []
        self.comboBox_milling_stage.currentIndexChanged.disconnect()
        self.comboBox_milling_stage.clear()
        self.comboBox_milling_stage.addItems([stage.name for stage in self.milling_stages])
        self.comboBox_milling_stage.currentIndexChanged.connect(lambda: self.update_milling_stage_ui())

        _remove_all_layers(self.viewer) # remove all shape layers

    def set_milling_stages(self, milling_stages: list[FibsemMillingStage]) -> None:
        logging.debug(f"Setting milling stages: {len(milling_stages)}")
        self.milling_stages = milling_stages
        
        # very explicitly set what is happening
        self.comboBox_milling_stage.currentIndexChanged.disconnect()
        self.comboBox_milling_stage.clear()
        self.comboBox_milling_stage.addItems([stage.name for stage in self.milling_stages])
        self.comboBox_milling_stage.currentIndexChanged.connect(lambda: self.update_milling_stage_ui())
        
        self.comboBox_patterns.currentIndexChanged.disconnect()
        self.comboBox_patterns.setCurrentText(self.milling_stages[0].pattern.name)
        self.comboBox_patterns.currentIndexChanged.connect(lambda: self.update_pattern_ui(None))
        
        logging.debug(f"Set milling stages: {len(milling_stages)}")
        self.update_milling_stage_ui()
        # self.update_ui()

    def get_milling_stages(self):
        return self.milling_stages

    def get_point_from_ui(self):

        point = Point(x=self.doubleSpinBox_centre_x.value() * constants.MICRO_TO_SI, 
                      y=self.doubleSpinBox_centre_y.value() * constants.MICRO_TO_SI)

        return point

    def update_milling_stage_ui(self):

        # get the selected milling stage
        current_index = self.comboBox_milling_stage.currentIndex()
        if current_index == -1:
            _remove_all_layers(self.viewer)
            return

        milling_stage: FibsemMillingStage = self.milling_stages[current_index]

        # set the milling settings
        self.set_milling_settings_ui(milling_stage.milling)

        # set the pattern protcol
        self.update_pattern_ui(milling_stage)

    def toggle_live_update(self):

        live_update = self.checkBox_live_update.isChecked()
        self.pushButton_update_pattern.setVisible(not live_update)

    def update_ui_pattern(self):

        if self._UPDATING_PATTERN:
            return

        if self.checkBox_live_update.isChecked():
            self.update_ui()

    def update_milling_stage_from_ui(self):
                
        # get current milling stage
        current_index = self.comboBox_milling_stage.currentIndex()

        if current_index == -1:
            msg = f"No milling stages defined, cannot draw patterns."
            logging.warning(msg)
            napari.utils.notifications.show_warning(msg)
            return

        milling_stage = self.milling_stages[current_index]

        # update milling settings
        milling_stage.milling = self.get_milling_settings_from_ui()

        # update pattern and define
        milling_stage.pattern = self.get_pattern_from_ui_v2()

        napari.utils.notifications.show_info(f"Updated {milling_stage.name}.")
        return 

    def open_path_dialog(self):

        file_path = _get_file_ui(msg="Select Bitmap File",_filter= "*bmp")
        
        self.path_edit.setText(file_path)

    def update_pattern_ui(self,milling_stage: FibsemMillingStage = None):

        self._UPDATING_PATTERN = True

        # get current pattern
        if milling_stage is None:
            current_pattern_text = self.comboBox_patterns.currentText()
            patterns_available = [pattern.name for pattern in patterning.__PATTERNS__]
            pattern_available_index = patterns_available.index(current_pattern_text)
            pattern = patterning.__PATTERNS__[pattern_available_index]
            pattern_protocol = self.protocol["patterns"][pattern.name]
            point = None
        else:
            pattern = milling_stage.pattern
            pattern_protocol = milling_stage.pattern.protocol
            point = milling_stage.pattern.point
            self.comboBox_patterns.currentIndexChanged.disconnect()
            self.comboBox_patterns.setCurrentText(milling_stage.pattern.name)
            self.comboBox_patterns.currentIndexChanged.connect(lambda: self.update_pattern_ui(None))

        logging.debug(f"PATTERN: {pattern.name}, PROTOCOL: {pattern_protocol}")

        # clear layout
        for i in reversed(range(self.gridLayout_patterns.count())):
            self.gridLayout_patterns.itemAt(i).widget().setParent(None)

        # set widgets for each required key / value
        for i, key in enumerate(pattern.required_keys):
            if key == "path":
                label = QtWidgets.QLabel(key)
                self.path_edit = QtWidgets.QLineEdit()
                self.gridLayout_patterns.addWidget(label, i, 0)
                self.gridLayout_patterns.addWidget(self.path_edit, i, 1)
                self.path_edit.setText(pattern_protocol[key])
                path_explorer = QtWidgets.QPushButton("...")
                self.gridLayout_patterns.addWidget(path_explorer, i, 2)
                path_explorer.clicked.connect(self.open_path_dialog)
                self.path_edit.editingFinished.connect(self.update_ui_pattern)
                continue
            
            if key == "scan_direction":
                label = QtWidgets.QLabel(key)
                self.comboBox_scan_direction = QtWidgets.QComboBox()
                self.gridLayout_patterns.addWidget(label, i, 0)
                self.gridLayout_patterns.addWidget(self.comboBox_scan_direction, i, 1)
                self.comboBox_scan_direction.addItems(self._available_scan_directions)
                scan_direction_to_use = pattern_protocol[key] if pattern_protocol[key] in self._available_scan_directions else self._available_scan_directions[0]
                # logging.info(f'Scan direction to use: {scan_direction_to_use}, available: {self._available_scan_directions}')
                self.comboBox_scan_direction.setCurrentText(scan_direction_to_use)
                self.comboBox_scan_direction.currentIndexChanged.connect(self.update_ui_pattern)
                continue

            if key == "cleaning_cross_section":
                if isinstance(self.microscope, ThermoMicroscope) and pattern.name == "Circle":
                    continue
                label = QtWidgets.QLabel(key)
                self.checkbox_cleaning_cross_section = QtWidgets.QCheckBox()
                self.gridLayout_patterns.addWidget(label, i, 0)
                self.gridLayout_patterns.addWidget(self.checkbox_cleaning_cross_section, i, 1)
                self.checkbox_cleaning_cross_section.setChecked(pattern_protocol[key])
                self.checkbox_cleaning_cross_section.stateChanged.connect(self.update_ui_pattern)
                continue

            if key == "inverted":
                label = QtWidgets.QLabel(key)
                self.checkbox_inverted = QtWidgets.QCheckBox()
                self.gridLayout_patterns.addWidget(label, i, 0)
                self.gridLayout_patterns.addWidget(self.checkbox_inverted, i, 1)
                self.checkbox_inverted.setChecked(pattern_protocol[key])
                self.checkbox_inverted.stateChanged.connect(self.update_ui_pattern)
                continue

            if key == "cross_section":
                label = QtWidgets.QLabel(key)
                self.comboBox_cross_section = QtWidgets.QComboBox()
                self.gridLayout_patterns.addWidget(label, i, 0)
                self.gridLayout_patterns.addWidget(self.comboBox_cross_section, i, 1)               
                self.comboBox_cross_section.addItems([section.name for section in CrossSectionPattern])
                self.comboBox_cross_section.setCurrentText(pattern_protocol.get(key, "Rectangle"))
                self.comboBox_cross_section.currentIndexChanged.connect(self.update_ui_pattern)
                continue

            # limits
            min_val = -1000 if key in _LINE_KEYS else 0

            label = QtWidgets.QLabel(key)
            spinbox = QtWidgets.QDoubleSpinBox()
            spinbox.setDecimals(3)
            spinbox.setSingleStep(0.001)
            spinbox.setRange(min_val, 1000)
            spinbox.setValue(0)
            self.gridLayout_patterns.addWidget(label, i, 0)
            self.gridLayout_patterns.addWidget(spinbox, i, 1)
            spinbox.setKeyboardTracking(False)

            # get default values from self.protocol and set values
            if key in pattern_protocol:
                value = _scale_value(key, pattern_protocol[key], constants.SI_TO_MICRO)
                spinbox.setValue(value if value is not None else 0)
            
            spinbox.valueChanged.connect(self.update_ui_pattern)

        if point is not None:
            self.doubleSpinBox_centre_x.setValue(point.x * constants.SI_TO_MICRO)
            self.doubleSpinBox_centre_y.setValue(point.y * constants.SI_TO_MICRO)

        if self.milling_stages:
            self.update_ui()

        self._UPDATING_PATTERN = False
    
    def get_pattern_settings_from_ui(self, pattern: patterning.BasePattern):
        # get pattern protocol from ui
        pattern_dict = {}
        for i, key in enumerate(pattern.required_keys):
            if key == "path":
                # add path in ui and get from there
                path = self.path_edit.text()
                pattern_dict[key] = path
                continue
            if key == "scan_direction":
                pattern_dict[key] = self.comboBox_scan_direction.currentText()
                continue
            if key == "cleaning_cross_section":
                if isinstance(self.microscope, ThermoMicroscope) and pattern.name == "Circle":
                    continue
                pattern_dict[key] = self.checkbox_cleaning_cross_section.isChecked()
                continue
            if key == "inverted":
                pattern_dict[key] = self.checkbox_inverted.isChecked()
                continue
            if key == "cross_section":
                pattern_dict[key] = self.comboBox_cross_section.currentText()
                continue
                
            spinbox = self.gridLayout_patterns.itemAtPosition(i, 1).widget()
            value = _scale_value(key, spinbox.value(), constants.MICRO_TO_SI)
            # value = value * constants.DEGREES_TO_RADIANS if key in _ANGLE_KEYS else value
            pattern_dict[key] = value # TODO: not everythign is in microns
        return pattern_dict

    def get_pattern_from_ui_v2(self):

        # get current pattern
        pattern = patterning.get_pattern(self.comboBox_patterns.currentText())

        pattern_dict = self.get_pattern_settings_from_ui(pattern)
        # define pattern
        point = Point(x=self.doubleSpinBox_centre_x.value() * constants.MICRO_TO_SI, y=self.doubleSpinBox_centre_y.value() * constants.MICRO_TO_SI)
        pattern.define(protocol=pattern_dict, point=point)
        
        return pattern

    def _single_click(self, layer, event):
        """Callback for single click on image layer."""
        if event.button != 1 or 'Shift' not in event.modifiers or self.milling_stages == []:
            return

        if not self._PATTERN_IS_MOVEABLE:
            msg = f"Pattern is not moveable."
            logging.info(msg)
            napari.utils.notifications.show_info(msg)
            return

        self._UPDATING_PATTERN = True

        # get coords
        coords = layer.world_to_data(event.position)

        # TODO: dimensions are mixed which makes this confusing to interpret... resolve
        coords, beam_type, image = self.image_widget.get_data_from_coord(coords)
        
        if beam_type is not BeamType.ION:
            napari.utils.notifications.show_info(
                f"Please right click on the {BeamType.ION.name} image to move pattern."
            )
            return
        
        point = conversions.image_to_microscope_image_coordinates(
                Point(x=coords[1], y=coords[0]), image.data, image.metadata.pixel_size.x,
            )
        
        clicked = deepcopy(point)

        # only move the pattern if milling widget is activate and beamtype is ion?
        renewed_patterns = []
        _patterns_valid = True

        current_stage_index = self.comboBox_milling_stage.currentIndex()

        diff = point - self.milling_stages[current_stage_index].pattern.point

        for idx, milling_stage in enumerate(self.milling_stages):
        # loop to check through all patterns to see if they are in bounds
            if 'Control' not in event.modifiers:
                if idx != current_stage_index: 
                    continue
            pattern_dict_existing = deepcopy(milling_stage.pattern.protocol)
            pattern_name = milling_stage.pattern.name
            pattern_renew = patterning.get_pattern(pattern_name)

            if self.checkBox_relative_move.isChecked():
                point = Point(x=milling_stage.pattern.point.x + diff.x, y=milling_stage.pattern.point.y + diff.y)
            
            # if the pattern is a line, we also need to update start_x, start_y, end_x, end_y to move with the click
            if pattern_name == "Line":
                pattern_dict_existing["start_x"] += diff.x
                pattern_dict_existing["start_y"] += diff.y
                pattern_dict_existing["end_x"] += diff.x
                pattern_dict_existing["end_y"] += diff.y
                # TODO: this doesnt work if the line is rotated at all
                
            pattern_renew.define(protocol=pattern_dict_existing, point=point)
            pattern_renew.point = point

            pattern_is_valid = self.valid_pattern_location(pattern_renew)
            # TODO: if the line goes out of bounds it is not reset correcectly, need to fix this

            if not pattern_is_valid:
                logging.warning(f"Could not move Patterns, out of bounds at at {point}")
                napari.utils.notifications.show_warning(f"Patterns is not within the image.")
                _patterns_valid = False
                break
            else:
                renewed_patterns.append(pattern_renew)

        _redraw = False

        if _patterns_valid:

            if len(renewed_patterns) == len(self.milling_stages):
                for milling_stage, pattern_renew in zip(self.milling_stages, renewed_patterns):
                    milling_stage.pattern = pattern_renew
                _redraw = True
            elif len(renewed_patterns)>0:
                self.milling_stages[current_stage_index].pattern = renewed_patterns[0]
                _redraw = True

        if _redraw:
            self.doubleSpinBox_centre_x.setValue(clicked.x * constants.SI_TO_MICRO)
            self.doubleSpinBox_centre_y.setValue(clicked.y * constants.SI_TO_MICRO) # THIS TRIGGERS AN UPDATE
            logging.info(f"Moved patterns to {point} ")

            # if current pattern is Line, we need to update the ui elements
            if self.milling_stages[current_stage_index].pattern.name == "Line":
                self.update_pattern_ui(milling_stage=self.milling_stages[current_stage_index])

            logging.debug({
                "msg": "move_milling_patterns",                                     # message type
                "pattern": self.milling_stages[current_stage_index].pattern.name,   # pattern name
                "dm": diff.to_dict(), # x, y                                        # metres difference 
                "beam_type": BeamType.ION.name,                                     # beam type
            })
            
            self.update_ui(milling_stages=self.milling_stages)
            self.milling_position_changed.emit()
    

        
        self._UPDATING_PATTERN = False

    def valid_pattern_location(self, stage_pattern: FibsemMillingStage) -> bool:

        if stage_pattern.name == "Fiducial":
            _,flag = _calculate_fiducial_area_v2(image=self.image_widget.ib_image, fiducial_centre = deepcopy(stage_pattern.point), fiducial_length = stage_pattern.patterns[0].height)
            
            if flag:
                napari.utils.notifications.show_warning(f"Fiducial reduce area is not within the image.")
                return False
            else:
                return True    
        
        for pattern_settings in stage_pattern.patterns:
            if isinstance(pattern_settings, patterning.CirclePattern):
                shape = convert_pattern_to_napari_circle(pattern_settings=pattern_settings, image=self.image_widget.ib_image)
            else:
                shape = convert_pattern_to_napari_rect(pattern_settings=pattern_settings, image=self.image_widget.ib_image)
            resolution = [self.image_widget.ib_image.data.shape[1], self.image_widget.ib_image.data.shape[0]]
            output = validate_pattern_placement(patterns=shape, resolution=resolution,shape=shape)
            if not output:
                return False
        
        return True
   
    def set_milling_settings_ui(self, milling: FibsemMillingSettings) -> None:

        # match to closest available milling current
        # idx = np.argmin(np.abs(np.array(self.AVAILABLE_MILLING_CURRENTS) - milling.milling_current))
        # print("milling current: ", milling.milling_current) # NB: this is not set at init TODO:
        self.doubleSpinBox_milling_current.setValue(milling.milling_current * constants.SI_TO_NANO)
        self.comboBox_application_file.setCurrentText(milling.application_file)
        self.doubleSpinBox_rate.setValue(milling.rate*constants.SI_TO_NANO)
        self.doubleSpinBox_dwell_time.setValue(milling.dwell_time * constants.SI_TO_MICRO)
        self.doubleSpinBox_spot_size.setValue(milling.spot_size * constants.SI_TO_MICRO)
        self.doubleSpinBox_hfw.setValue(milling.hfw * constants.SI_TO_MICRO)
        self.comboBox_preset.setCurrentText(str(milling.preset))
        self.spinBox_voltage.setValue(milling.milling_voltage)

    def get_milling_settings_from_ui(self):

        current_amps = float(self.doubleSpinBox_milling_current.value()) * constants.NANO_TO_SI

        milling_settings = FibsemMillingSettings(
            milling_current=current_amps,
            application_file=self.comboBox_application_file.currentText(),
            rate=self.doubleSpinBox_rate.value()*1e-9,
            dwell_time = self.doubleSpinBox_dwell_time.value() * constants.MICRO_TO_SI,
            spot_size=self.doubleSpinBox_spot_size.value() * constants.MICRO_TO_SI,
            hfw=self.doubleSpinBox_hfw.value() * constants.MICRO_TO_SI,
            preset= self.comboBox_preset.currentText(),
            spacing=self.doubleSpinBox_spacing.value(),
            milling_voltage=self.spinBox_voltage.value(),
        )

        return milling_settings

    def update_ui(self, milling_stages: list[FibsemMillingStage] = None):
        self.doubleSpinBox_hfw.setValue(self.image_widget.doubleSpinBox_image_hfw.value())

        if milling_stages is None and len(self.milling_stages) < 1:
            return
        
        t0 = time.time()
        if milling_stages is None:
            self.update_milling_stage_from_ui() # update milling stage from ui
            milling_stages = self.get_milling_stages() # get the latest milling stages from the ui

        t1 = time.time() 

        if not milling_stages:
            _remove_all_layers(self.viewer)
            return

        # make milling stage a list if it is not
        if not isinstance(milling_stages, list):
            milling_stages = [milling_stages]

        # # check hfw threshold
        for stage in milling_stages:
            if stage.pattern.name == "Trench":
                if stage.pattern.protocol["trench_height"] / stage.milling.hfw < cfg.MILL_HFW_THRESHOLD:
                    napari.utils.notifications.show_warning(f"Pattern dimensions are too small for milling. Please decrease the image hfw or increase the trench height.")
                    _remove_all_layers(self.viewer)
                    return

        t2 = time.time()
        try:
            draw_crosshair = self.checkBox_show_milling_crosshair.isChecked()
            from fibsem.structures import FibsemImage
            if not isinstance(self.image_widget.ib_image, FibsemImage):
                raise Exception(f"No Ion Image, cannot draw patterns. Please take an image.")
            if not isinstance(self.image_widget.eb_image, FibsemImage):
                raise Exception(f"No Electron Image, cannot draw patterns. Please take an image.") # TODO: this is unintuitive why this is required -> ui issue only
            # clear patterns then draw new ones
            _draw_patterns_in_napari(self.viewer, 
                ib_image=self.image_widget.ib_image, 
                eb_image=self.image_widget.eb_image, 
                milling_stages = milling_stages,
                draw_crosshair=draw_crosshair)  # TODO: add names and legend for this
        
        except Exception as e:
            napari.utils.notifications.show_error(f"Error drawing patterns: {e}")
            logging.error(e)
            return
        t2 = time.time()
        logging.debug(f"UPDATE_UI: GET: {t1-t0}, DRAW: {t2-t1}")
        self.viewer.layers.selection.active = self.image_widget.eb_layer

    def _toggle_interactions(self, enabled: bool = True, caller: str = None, milling: bool = False):

        """Toggle microscope and pattern interactions."""
        self.pushButton_update_pattern.setEnabled(enabled)
        self.pushButton_add_milling_stage.setEnabled(enabled)
        self.pushButton_remove_milling_stage.setEnabled(enabled)
        # self.pushButton_save_milling_stage.setEnabled(enabled)
        self.pushButton_run_milling.setEnabled(enabled)
        if enabled:
            self.pushButton_run_milling.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.pushButton_run_milling.setText("Run Milling")
            self.pushButton_update_pattern.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
            self.pushButton_add_milling_stage.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.pushButton_remove_milling_stage.setStyleSheet(_stylesheets._RED_PUSHBUTTON_STYLE)
        elif milling:
            self.pushButton_run_milling.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)
            self.pushButton_run_milling.setText("Running...")
        else:
            self.pushButton_run_milling.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_update_pattern.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_add_milling_stage.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_remove_milling_stage.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)

        if caller is None:
            self.parent.image_widget._toggle_interactions(enabled, caller="milling")
            self.parent.movement_widget._toggle_interactions(enabled, caller="milling")


    def run_milling(self):
        
        worker = self.run_milling_step()
        worker.finished.connect(self.run_milling_finished)
        worker.start()
        
    
    def start_progress_thread(self,info):

        est_time = info['estimated_time']  # TODO: add some extra time to this to account for changing currents etc
        idx = info['idx']
        total = info['total']

        self.progressBar_milling.setVisible(True)
        self.progressBar_milling.setValue(0)
        self.progressBar_milling.setStyleSheet("QProgressBar::chunk "
                          "{"
                          "background-color: green;"
                          "}")
        self.progressBar_milling.setFormat(f"Milling stage {idx+1} of {total}: {est_time:.1f}s")

        # info = [idx, total, est_time]

        self.progress_bar_worker = self.start_progress_bar(info)
        self.progress_bar_worker.finished.connect(self.finish_progress_bar)

        self.progress_bar_worker.start()

    @thread_worker
    def start_progress_bar(self,info):
        
        est_time = info['estimated_time']
        info['progress_percent'] = 0
 
        i = 0
        # time.sleep(2)
        inc = 0.5
        while i < est_time:
            time.sleep(inc)
            progress_percent = (i+inc)/est_time
            info['progress_percent'] = progress_percent
            info['est_time'] = est_time - i
            self._progress_bar_update.emit(info)
            i += inc 
            yield


    def update_progress_bar(self, info):
        
        value = info['progress_percent']
        idx = info['idx']
        total = info['total']
        est_time = info['est_time']

        self.progressBar_milling.setVisible(True)
        self.progressBar_milling.setValue(value*100)
        self.progressBar_milling.setFormat(f"Milling Stage {idx+1}/{total}: {est_time:.1f}s remaining...")

    def finish_progress_bar(self):
        self.progressBar_milling.setVisible(False)
        self.progressBar_milling.setValue(0)

    def _quit_progress_bar(self):
        self.progress_bar_worker.quit()

    @thread_worker
    def run_milling_step(self):

        milling_stages = self.get_milling_stages()
        self._toggle_interactions(enabled=False,milling=True)
        
        # TODO: enable new milling after testing
        USE_MILL_STAGES_UI_V2 = False
        if USE_MILL_STAGES_UI_V2:
            milling.mill_stages_ui(microscope=self.microscope, 
                                stages=milling_stages, 
                                parent_ui=self)
            return
            
        for idx,stage in enumerate(milling_stages):
            self.milling_notification.emit(f"Preparing: {stage.name}")
            if stage.pattern is not None:
                log_status_message(stage, f"RUNNING_MILLING_STAGE_{stage.name}") # TODO: refactor to json
                log_status_message(stage, f"MILLING_PATTERN_{stage.pattern.name}: {stage.pattern.patterns}")
                log_status_message(stage, f"MILLING_SETTINGS_{stage.milling}")
                try:
                    milling.setup_milling(self.microscope, mill_settings=stage.milling)

                    microscope_patterns = milling.draw_patterns(self.microscope, stage.pattern.patterns)
                    estimated_time = milling.estimate_milling_time(self.microscope, microscope_patterns)
                    progress_bar_dict = {"estimated_time": estimated_time, "idx": idx, "total": len(milling_stages)}
                    self._progress_bar_start.emit(progress_bar_dict)

                    self.milling_notification.emit(f"Running {stage.name}...")
                    milling.run_milling(self.microscope, stage.milling.milling_current, stage.milling.milling_voltage)

                    # TODO implement a special case for overtilt milling

                except Exception as e:
                    napari.utils.notifications.show_error(f"Error running milling stage: {stage.name}")
                    logging.error(e)
                finally:
                    milling.finish_milling(self.microscope, 
                                           imaging_current=self.microscope.system.ion.beam.beam_current, 
                                           imaging_voltage=self.microscope.system.ion.beam.voltage)

                log_status_message(stage, "MILLING_COMPLETED_SUCCESSFULLY")
                self._progress_bar_quit.emit()

            self.milling_notification.emit(f"Milling stage complete: {stage.name}")
        self.milling_notification.emit(f"Milling complete. {len(self.milling_stages)} stages completed.")


    def update_milling_ui(self, msg: str):
        logging.info(msg)
        napari.utils.notifications.notification_manager.records.clear()
        napari.utils.notifications.show_info(msg)

    def run_milling_finished(self):

        # take new images and update ui
        self._toggle_interactions(enabled=True)
        self.image_widget.take_reference_images()
        self.update_ui()
        self._milling_finished.emit()
        self._quit_progress_bar()
        self.finish_progress_bar()




def main():
    millings_stages = [
    FibsemMillingStage(
        name="Milling Stage X",
        num = 1,
        milling = FibsemMillingSettings(hfw=400e-6),
        pattern = patterning.get_pattern("Trench"),
    ),
        FibsemMillingStage(
        name="Milling Stage 2",
        num = 2,
        milling = FibsemMillingSettings(hfw=200e-6),
        pattern = patterning.get_pattern("Horseshoe"),
    )
    ]
    viewer = napari.Viewer(ndisplay=2)
    movement_widget = FibsemMillingWidget()
    viewer.window.add_dock_widget(
        movement_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()


from typing import Optional
import logging

import numpy as np
import torch
import torch.nn.functional as F

import segmentation_models_pytorch as smp
from fibsem.segmentation.utils import decode_segmap, download_checkpoint

from pathlib import Path
import os



# NOTE: these models contain numeric training bugs that were fixed in later versions. For reproducibility, we need to re-implement the bug for these models...
# please contact (pat) if these models arent working as expected. 
__DEPRECIATED_CHECKPOINTS__ = [
    "autolamella-02-34.pt",
    "autolamella-03-34.pt",
    "autolamella-04-34.pt",
    "autolamella-05-34.pt",
    "autolamella-latest.pt",
    "openfibsem-01-18.pt",
    "openfibsem-02-18.pt",
    "openfibsem-03-18.pt",
    "openfibsem-baseline-34.pt"
    "openfibsem-baseline-latest.pt",
    "autoliftout-serial-01-34.pt",
]


class SegmentationModel:
    def __init__(
        self,
        checkpoint: str = None,
        encoder: str = "resnet34",
        mode: str = "eval",
        num_classes: int = 3,
        _fix_numeric_scaling: bool = True,
        use_v2: bool = True,
    ) -> None:
        super().__init__()

        self.checkpoint: str = checkpoint
        self.mode = mode
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.num_classes = num_classes
        self._fix_numeric_scaling = _fix_numeric_scaling

        if checkpoint in __DEPRECIATED_CHECKPOINTS__:
            self._fix_numeric_scaling = False

        if use_v2:
            self.model = self.load_model_v2(checkpoint=checkpoint)
        else:
            self.load_model(checkpoint=checkpoint, encoder=encoder)

    # TODO: deprecate this fully
    def load_model(self, checkpoint: Optional[str], encoder: str = "resnet18") -> None:
        """Load the model, and optionally load a checkpoint"""

        # show depreciation warning
        print(f"WARNING: {checkpoint} is a depreciated checkpoint. Please use the latest checkpoint instead.")

        self.model = smp.Unet(
            encoder_name=encoder,
            encoder_weights="imagenet",
            in_channels=1,  # grayscale images
            classes=self.num_classes,
        )
        self.model.to(self.device)        
        
        
        self.load_weights(checkpoint=checkpoint)
        if checkpoint:
            checkpoint = download_checkpoint(checkpoint)

            self.checkpoint = checkpoint
            checkpoint_state = torch.load(checkpoint, map_location=self.device)
            self.model.load_state_dict(checkpoint_state)
        if self._fix_numeric_scaling:
            self.model.eval() # this causes a bug? why -> input needs to be scaled between 0-1
        if self.mode == "train":
            self.model.train()

    def load_model_v2(self, checkpoint: str):

        # download checkpoint if needed
        checkpoint = download_checkpoint(checkpoint)
        checkpoint_dict = torch.load(checkpoint, map_location=self.device)

        self.model = smp.Unet(
            encoder_name=checkpoint_dict["encoder"],
            encoder_weights="imagenet",
            in_channels=1,  # grayscale images
            classes=checkpoint_dict["nc"],
        )
        self.model.to(self.device)

        self.model.load_state_dict(checkpoint_dict["checkpoint"])

        if self._fix_numeric_scaling:
            self.model.eval() # this causes a bug? why -> input needs to be scaled between 0-1
        if self.mode == "train":
            self.model.train()

        # metadata
        self.checkpoint = os.path.basename(checkpoint)
        self.num_classes = checkpoint_dict["nc"]
        self.encoder = checkpoint_dict["encoder"]

        return self.model


    def pre_process(self, img: np.ndarray) -> torch.Tensor:
        """Pre-process the image for inference"""
        img_t = torch.Tensor(img).float().to(self.device)

        if self._fix_numeric_scaling:
            img_t /=  255.0 # scale float to 0 - 1
        if img_t.ndim == 2:
            img_t = img_t.unsqueeze(0).unsqueeze(0)  # add batch dim and channel dim
        elif img_t.ndim == 3:
            if img_t.shape[0] > 1:  # means the first dim is batch dim
                img_t = img_t.unsqueeze(1)  # add channel dim
            else:
                img_t = img_t.unsqueeze(0)  # add batch dim

        assert img_t.ndim == 4, f"Expected 4 dims, got {img_t.ndim}"

        logging.debug({"msg": "pre_process", "shape": img_t.shape, "dtype": img_t.dtype, "min": img_t.min(), "max": img_t.max()})

        return img_t

    def inference(self, img: np.ndarray, rgb: bool = True) -> np.ndarray:
        """Run model inference on the input image"""
        with torch.no_grad():
            img_t = self.pre_process(img)

            outputs = self.model(img_t)
            outputs = F.softmax(outputs, dim=1)
            masks = torch.argmax(outputs, dim=1).detach().cpu().numpy()
        
        # decode to rgb
        if rgb:
            masks = self.postprocess(masks, nc=self.num_classes)

        # TODO: return masks, scores, logits
        return masks

    def inference_v2(self, img: np.ndarray, rgb: bool = True) -> np.ndarray:
        """Run model inference on the input image"""
        with torch.no_grad():
            img_t = self.pre_process(img)

            outputs = self.model(img_t)
            outputs = F.softmax(outputs, dim=1)
            masks = torch.argmax(outputs, dim=1).detach().cpu().numpy()
        
        # decode to rgb
        if rgb:
            masks = self.postprocess(masks, nc=self.num_classes)

        # TODO: return masks, scores, logits
        return masks, outputs
        
    def postprocess(self, masks, nc):
        # TODO: vectorise this properly
        # TODO: use decode_segmap_v2
        output_masks = []
        for i in range(len(masks)):
            output_masks.append(decode_segmap(masks[i], nc=nc))

        if len(output_masks) == 1:
            output_masks = output_masks[0]
        return np.array(output_masks)


def get_backend(checkpoint: str) -> str:

    if "nnunet" in checkpoint:
        return "nnunet"
    elif "onnx" in checkpoint:
        return "onnx"
    else:
        return "smp"

def load_model(
    checkpoint: Path, encoder: str = "resnet34", nc: int = 3, _fix_numeric_scaling: bool = True, backend = None
) -> SegmentationModel:
    """Load a model checkpoint
    backend: str, optional The backend to use. If None, will try to infer from the checkpoint name"""
    
    if backend is None:
        backend = get_backend(checkpoint=checkpoint)

    # load model
    if backend == "nnunet":
        from fibsem.segmentation.nnunet_model import SegmentationModelNNUnet
        model = SegmentationModelNNUnet(checkpoint=checkpoint)
    elif backend == "onnx":
        from fibsem.segmentation.onnx_model import SegmentationModelONNX
        model = SegmentationModelONNX(checkpoint=checkpoint)
    else:
        model = SegmentationModel(checkpoint=checkpoint, encoder=encoder, num_classes=nc, _fix_numeric_scaling=_fix_numeric_scaling)

    return model


if __name__ == "__main__":

    model = SegmentationModel(checkpoint="checkpoint_train.pth.tar", mode="train")



import os

import fibsem
import napari
import napari.utils.notifications
from autoscript_sdb_microscope_client import SdbMicroscopeClient
from fibsem import calibration, constants, utils
from fibsem.structures import BeamType, GammaSettings, ImageSettings, MultiChemSettings
from fibsem.ui.qtdesigner_files import FibsemMultiChemWidget
from PyQt5 import QtWidgets


class FibsemMultiChemWidget(FibsemMultiChemWidget.Ui_Form, QtWidgets.QWidget):
    def __init__(
        self,
        microscope: SdbMicroscopeClient = None,
        mc_settings: MultiChemSettings = None,
        parent=None,
    ):
        super(FibsemMultiChemWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.microscope = microscope

        self.setup_connections()
        if mc_settings is not None:
            self.update_ui_from_settings(mc_settings)

    def setup_connections(self):

        print("setup connections")

        self.comboBox_beam_type.addItems([beam.name for beam in BeamType])
        self.comboBox_application_file.addItems(["cryo_Pt_dep"])
        self.comboBox_gas.addItems(["Pt cryo"])
        self.comboBox_position.addItems(["cryo"])

        self.pushButton_run_sputtering.clicked.connect(self.run_sputtering)

    def update_ui_from_settings(self, mc_settings: MultiChemSettings):

        self.comboBox_beam_type.setCurrentText(mc_settings.beam_type.name)
        self.comboBox_application_file.setCurrentText(mc_settings.application_file)
        self.comboBox_gas.setCurrentText(mc_settings.gas)
        self.comboBox_position.setCurrentText(mc_settings.position)
        self.doubleSpinBox_hfw.setValue(mc_settings.hfw * constants.SI_TO_MICRO)
        self.doubleSpinBox_length.setValue(mc_settings.length * constants.SI_TO_MICRO)
        self.doubleSpinBox_time.setValue(mc_settings.time)

    def get_settings_from_ui(self):

        mc_settings = MultiChemSettings(
            application_file=self.comboBox_application_file.currentText(),
            gas=self.comboBox_gas.currentText(),
            position=self.comboBox_position.currentText(),
            beam_type=BeamType[self.comboBox_beam_type.currentText()],
            hfw=self.doubleSpinBox_hfw.value() * constants.MICRO_TO_SI,
            length=self.doubleSpinBox_length.value() * constants.MICRO_TO_SI,
            time=self.doubleSpinBox_time.value(),
        )

        print(f"get settings: {mc_settings}")

        return mc_settings

    def run_sputtering(self):

        mc_settings = self.get_settings_from_ui()

        print("run sputtering")

        from fibsem import utils

        # utils.sputter_multichem_v2(self.microscope, multichem_settings)


def main():

    viewer = napari.Viewer(ndisplay=2)
    multichem_widget = FibsemMultiChemWidget(mc_settings=MultiChemSettings())
    viewer.window.add_dock_widget(
        multichem_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()
